import os
import sys

# Get the absolute path of the current script file

script_path = os.path.abspath(__file__)

# Get the root directory by going up one level from the script directory
project_root = os.path.dirname(os.path.dirname(script_path))

sys.path.insert(0, project_root)
sys.path.insert(0, os.getcwd())  # Add the current directory as well

# for testing pyside6
# os.environ['QT_API'] = 'pyside6'

# for testing pyqt6
# os.environ['QT_API'] = 'pyqt6'

from PySide6.QtGui import QFont, QIcon, QPixmap
from PySide6.QtWidgets import QApplication, \
    QSplashScreen
from PySide6.QtSql import QSqlDatabase

from pyqt_openai.config_loader import CONFIG_MANAGER

from pyqt_openai.mainWindow import MainWindow
from pyqt_openai.util.script import handle_exception
from pyqt_openai.updateSoftwareDialog import update_software
from pyqt_openai.sqlite import get_db_filename

from pyqt_openai import DEFAULT_APP_ICON


# Application
class App(QApplication):
    def __init__(self, *args):
        super().__init__(*args)
        self.setQuitOnLastWindowClosed(False)
        self.setWindowIcon(QIcon(DEFAULT_APP_ICON))
        self.splash = QSplashScreen(QPixmap(DEFAULT_APP_ICON))
        self.splash.show()

        self.__initQSqlDb()
        self.__initFont()

        self.__showMainWindow()
        self.splash.finish(self.main_window)

        update_software()

    def __initQSqlDb(self):
        # Set up the database and table model (you'll need to configure this part based on your database)
        self.__db = QSqlDatabase.addDatabase('QSQLITE')
        self.__db.setDatabaseName(get_db_filename())
        self.__db.open()

    def __initFont(self):
        font_family = CONFIG_MANAGER.get_general_property('font_family')
        font_size = CONFIG_MANAGER.get_general_property('font_size')
        QApplication.setFont(QFont(font_family, font_size))

    def __showMainWindow(self):
        self.main_window = MainWindow()
        self.main_window.show()

# Set the global exception handler
sys.excepthook = handle_exception

def main():
    app = App(sys.argv)
    sys.exit(app.exec())


if __name__ == "__main__":
    main()


from PySide6.QtWidgets import QLabel, QFormLayout, QGroupBox, QDialog

from pyqt_openai import DEFAULT_SHORTCUT_FIND_PREV, DEFAULT_SHORTCUT_FIND_NEXT, DEFAULT_SHORTCUT_PROMPT_BEGINNING, \
    DEFAULT_SHORTCUT_PROMPT_ENDING, DEFAULT_SHORTCUT_SUPPORT_PROMPT_COMMAND, DEFAULT_SHORTCUT_FULL_SCREEN, \
    DEFAULT_SHORTCUT_FIND, DEFAULT_SHORTCUT_JSON_MODE, DEFAULT_SHORTCUT_LEFT_SIDEBAR_WINDOW, \
    DEFAULT_SHORTCUT_RIGHT_SIDEBAR_WINDOW, DEFAULT_SHORTCUT_CONTROL_PROMPT_WINDOW, DEFAULT_SHORTCUT_SETTING, \
    DEFAULT_SHORTCUT_SEND, DEFAULT_SHORTCUT_SHOW_TOOLBAR, DEFAULT_SHORTCUT_SHOW_SECONDARY_TOOLBAR, DEFAULT_SHORTCUT_FOCUS_MODE
from pyqt_openai.lang.translations import LangClass


class ShortcutDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.__shortcuts = {
            'SHORTCUT_FIND_PREV': {'label': f"{LangClass.TRANSLATIONS['Find']} - {LangClass.TRANSLATIONS['Previous']}", 'value': DEFAULT_SHORTCUT_FIND_PREV},
            'SHORTCUT_FIND_NEXT': {'label': f"{LangClass.TRANSLATIONS['Find']} - {LangClass.TRANSLATIONS['Next']}", 'value': DEFAULT_SHORTCUT_FIND_NEXT},
            'SHORTCUT_PROMPT_BEGINNING': {'label': LangClass.TRANSLATIONS['Prompt Beginning'], 'value': DEFAULT_SHORTCUT_PROMPT_BEGINNING},
            'SHORTCUT_PROMPT_ENDING': {'label': LangClass.TRANSLATIONS['Prompt Ending'], 'value': DEFAULT_SHORTCUT_PROMPT_ENDING},
            'SHORTCUT_SUPPORT_PROMPT_COMMAND': {'label': LangClass.TRANSLATIONS['Support Prompt Command'], 'value': DEFAULT_SHORTCUT_SUPPORT_PROMPT_COMMAND},
            'SHORTCUT_SHOW_TOOLBAR': {'label': LangClass.TRANSLATIONS['Show Toolbar'], 'value': DEFAULT_SHORTCUT_SHOW_TOOLBAR},
            'SHOW_SECONDARY_TOOLBAR': {'label': LangClass.TRANSLATIONS['Show Secondary Toolbar'], 'value': DEFAULT_SHORTCUT_SHOW_SECONDARY_TOOLBAR },
            'SHORTCUT_FOCUS_MODE': {'label': LangClass.TRANSLATIONS['Focus Mode'], 'value': DEFAULT_SHORTCUT_FOCUS_MODE},
            'SHORTCUT_FULL_SCREEN': {'label': LangClass.TRANSLATIONS['Full Screen'], 'value': DEFAULT_SHORTCUT_FULL_SCREEN},
            'SHORTCUT_FIND': {'label': LangClass.TRANSLATIONS['Find'], 'value': DEFAULT_SHORTCUT_FIND},
            'SHORTCUT_JSON_MODE': {'label': LangClass.TRANSLATIONS['JSON Mode'], 'value': DEFAULT_SHORTCUT_JSON_MODE},
            'SHORTCUT_LEFT_SIDEBAR_WINDOW': {'label': LangClass.TRANSLATIONS['Left Sidebar Window'], 'value': DEFAULT_SHORTCUT_LEFT_SIDEBAR_WINDOW},
            'SHORTCUT_RIGHT_SIDEBAR_WINDOW': {'label': LangClass.TRANSLATIONS['Right Sidebar Window'], 'value': DEFAULT_SHORTCUT_RIGHT_SIDEBAR_WINDOW},
            'SHORTCUT_CONTROL_PROMPT_WINDOW': {'label': LangClass.TRANSLATIONS['Control Prompt Window'], 'value': DEFAULT_SHORTCUT_CONTROL_PROMPT_WINDOW},
            'SHORTCUT_SETTING': {'label': LangClass.TRANSLATIONS['Setting'], 'value': DEFAULT_SHORTCUT_SETTING},
            'SHORTCUT_SEND': {'label': LangClass.TRANSLATIONS['Send'], 'value': DEFAULT_SHORTCUT_SEND},
        }
        self.__initUi()

    def __initUi(self):
        self.setWindowTitle(LangClass.TRANSLATIONS['Shortcuts'])

        lay = QFormLayout()

        shortcutGroupBox = QGroupBox(LangClass.TRANSLATIONS['Shortcuts'])
        shortcutGroupBox.setLayout(lay)

        for key, shortcut in self.__shortcuts.items():
            lineEdit = QLabel()
            lineEdit.setText(shortcut['value'])
            shortcut['lineEdit'] = lineEdit
            lay.addRow(shortcut['label'], lineEdit)

        self.setLayout(lay)


"""
This is the file that contains the global variables that are used, or possibly used, throughout the application.
"""

import anthropic
import google.generativeai as genai
from openai import OpenAI
from g4f.client import Client

from pyqt_openai import DEFAULT_GEMINI_MODEL, LLAMA_REQUEST_URL
from pyqt_openai.sqlite import SqliteDatabase
from pyqt_openai.util.llamapage_script import GPTLLamaIndexWrapper
from pyqt_openai.util.replicate_script import ReplicateWrapper

DB = SqliteDatabase()

LLAMAINDEX_WRAPPER = GPTLLamaIndexWrapper()

G4F_CLIENT = Client()

OPENAI_CLIENT = OpenAI(api_key='')
GEMINI_CLIENT = genai.GenerativeModel(DEFAULT_GEMINI_MODEL)
CLAUDE_CLIENT = anthropic.Anthropic(api_key='')
LLAMA_CLIENT = OpenAI(
    api_key='',
    base_url=LLAMA_REQUEST_URL
)

REPLICATE_CLIENT = ReplicateWrapper(api_key='')



"""
This file is used to store the constants and the global constants / variables that are used throughout the application.
Constants/Variables that are stored here are used throughout the application for the following purposes:
- Initial values related to the environment settings
- Values used internally within the application (e.g., application name, DB name, table name, etc.)
- Values related to the design and UI of the application
- Default LLM list for the application settings

Constants/Variables that are stored here are not supposed to be changed during the runtime except for __init__.py.
Variables which are used globally and can be changed are stored in globals.py.
"""

import json
import os
import shutil
import sys

import tomllib  # Python 3.11 built-in module
from pathlib import Path

# Load the pyproject.toml file

SRC_DIR = Path(__file__).resolve().parent # VividNode/pyqt_openai
ROOT_DIR = SRC_DIR.parent # VividNode
SETUP_FILENAME = ROOT_DIR / "pyproject.toml"

# Read the TOML file using tomllib (Python 3.11+)
with open(SETUP_FILENAME, "rb") as file:
    pyproject_data = tomllib.load(file)

# For the sake of following the PEP8 standard, we will declare module-level dunder names.
# PEP8 standard about dunder names: https://peps.python.org/pep-0008/#module-level-dunder-names

__version__ = pyproject_data["project"]["version"]
__author__ = pyproject_data["project"]["authors"][0]['name']

# Constants
# ----------------------------
# APP
PACKAGE_NAME = pyproject_data["project"]["name"]
OWNER = 'yjg30737'

DEFAULT_APP_NAME = 'VividNode'

AUTOSTART_REGISTRY_KEY = r"Software\Microsoft\Windows\CurrentVersion\Run"

# Check if the application is frozen (compiled with PyInstaller)
# If this is main.py, it will return False, that means it is not frozen.
def is_frozen():
    return hasattr(sys, 'frozen')

# The executable path of the application
def get_executable_path():
    if is_frozen():  # For PyInstaller
        executable_path = sys._MEIPASS
    else:
        executable_path = os.path.dirname(os.path.abspath(__file__))
    return executable_path

EXEC_PATH = get_executable_path()

# The current filename of the application
CURRENT_FILENAME = os.path.join(EXEC_PATH, f'{DEFAULT_APP_NAME}.exe')

def get_config_directory():
    if os.name == 'nt':  # Windows
        config_dir = os.path.join(os.getenv('APPDATA'), DEFAULT_APP_NAME)
    elif os.name == 'posix':  # macOS/Linux
        config_dir = os.path.join(os.getenv('XDG_CONFIG_HOME', os.path.expanduser('~/.config')), DEFAULT_APP_NAME)
    else:
        config_dir = os.path.expanduser(f'~/.{DEFAULT_APP_NAME}')  # Fallback

    # Ensure the directory exists
    if not os.path.exists(config_dir):
        os.makedirs(config_dir)

    return config_dir

UPDATE_DIR = get_config_directory()

# The default updater path (relative to the application's root directory)
UPDATER_PATH = os.path.join(UPDATE_DIR, 'Updater.exe')

# Move the Updater.exe to the config folder
def move_updater():
    original_updater_path = os.path.join(ROOT_DIR, 'Updater.exe')
    if os.path.exists(original_updater_path):
        if os.path.exists(UPDATER_PATH):
            os.remove(UPDATER_PATH)
        shutil.move(original_updater_path, UPDATER_PATH)

move_updater()

CONTACT = pyproject_data["project"]["authors"][0]['email']
APP_INITIAL_WINDOW_SIZE = (1280, 768)

TRANSPARENT_RANGE = 20, 100
TRANSPARENT_INIT_VAL = 100

LARGE_LABEL_PARAM = ('Arial', 32)
MEDIUM_LABEL_PARAM = ('Arial', 16)
SMALL_LABEL_PARAM = ('Arial', 10)

LICENSE = pyproject_data["project"]["license"]['text']
LICENSE_URL = 'https://github.com/yjg30737/pyqt-openai/blob/main/LICENSE'
KOFI_URL = 'https://ko-fi.com/junggyuyoon'
PAYPAL_URL = 'https://paypal.me/yjg30737'
GITHUB_URL = 'https://github.com/yjg30737/pyqt-openai'
DISCORD_URL = 'https://discord.gg/cHekprskVE'

QUICKSTART_MANUAL_URL = 'https://medium.com/@yjg30737/what-is-vividnode-how-to-use-it-4d8a9269a3c0'
LLAMAINDEX_URL = 'https://medium.com/@yjg30737/what-is-llamaindex-9b821d66568f'
HOW_TO_GET_OPENAI_API_KEY_URL = 'https://medium.com/@yjg30737/how-to-get-your-openai-api-key-e2193850932e'
HOW_TO_EXPORT_CHATGPT_CONVERSATION_HISTORY_URL = 'https://medium.com/@yjg30737/how-to-export-your-chatgpt-conversation-history-caa0946d6349'
HOW_TO_REPLICATE = 'https://medium.com/@yjg30737/10a2cb983ceb'
HOW_TO_GET_CLAUDE_API_KEY_URL = 'https://medium.com/@yjg30737/how-to-get-started-with-claude-api-step-by-step-guide-92b5e35ae0a0'
HOW_TO_GET_GEMINI_API_KEY_URL = 'https://medium.com/@yjg30737/e61a84d64c69'
HOW_TO_GET_LLAMA_API_KEY_URL = 'https://medium.com/@yjg30737/0dc1900e3606'

COLUMN_TO_EXCLUDE_FROM_SHOW_HIDE_CHAT = ['id']
COLUMN_TO_EXCLUDE_FROM_SHOW_HIDE_IMAGE = ['id', 'data']
DEFAULT_LANGUAGE = 'en_US'
LANGUAGE_DICT = {
    "English": "en_US",
    "Spanish": "es_ES",
    "Chinese": "zh_CN",
    "Russian": "ru_RU",
    "Korean": "ko_KR",
    "French": "fr_FR",
    "German": "de_DE",
    "Italian": "it_IT",
    "Hindi": "hi_IN",
    "Arabic": "ar_AE",
    "Japanese": "ja_JP",
    "Bengali": "bn_IN",
    "Urdu": "ur_PK",
    "Indonesian": "id_ID",
    "Portuguese": "pt_BR"
}

MESSAGE_PADDING = 16
MESSAGE_MAXIMUM_HEIGHT = 800
MESSAGE_MAXIMUM_HEIGHT_RANGE = 300, 1000

CONTEXT_DELIMITER = '\n'*2
PROMPT_IMAGE_SCALE = 200, 200
TOAST_DURATION = 3

## ICONS
ICON_PATH = os.path.join(EXEC_PATH, 'ico')

DEFAULT_APP_ICON = os.path.join(EXEC_PATH, 'icon.ico')

ICON_ADD = os.path.join(ICON_PATH, 'add.svg')
ICON_CASE = os.path.join(ICON_PATH, 'case.svg')
ICON_CLOSE = os.path.join(ICON_PATH, 'close.svg')
ICON_COPY = os.path.join(ICON_PATH, 'copy.svg')
ICON_CUSTOMIZE = os.path.join(ICON_PATH, 'customize.svg')
ICON_DELETE = os.path.join(ICON_PATH, 'delete.svg')
ICON_DISCORD = os.path.join(ICON_PATH, 'discord.svg')
ICON_EXPORT = os.path.join(ICON_PATH, 'export.svg')
ICON_FAVORITE_NO = os.path.join(ICON_PATH, 'favorite_no.svg')
ICON_FAVORITE_YES = os.path.join(ICON_PATH, 'favorite_yes.svg')
ICON_FOCUS_MODE = os.path.join(ICON_PATH, 'focus_mode.svg')
ICON_FULLSCREEN = os.path.join(ICON_PATH, 'fullscreen.svg')
ICON_GITHUB = os.path.join(ICON_PATH, 'github.svg')
ICON_HELP = os.path.join(ICON_PATH, 'help.svg')
ICON_HISTORY = os.path.join(ICON_PATH, 'history.svg')
ICON_IMPORT = os.path.join(ICON_PATH, 'import.svg')
ICON_INFO = os.path.join(ICON_PATH, 'info.svg')
ICON_NEXT = os.path.join(ICON_PATH, 'next.svg')
ICON_OPENAI = os.path.join(ICON_PATH, 'openai.png')
ICON_PREV = os.path.join(ICON_PATH, 'prev.svg')
ICON_PROMPT = os.path.join(ICON_PATH, 'prompt.svg')
ICON_QUESTION = os.path.join(ICON_PATH, 'question.svg')
ICON_REFRESH = os.path.join(ICON_PATH, 'refresh.svg')
ICON_REGEX = os.path.join(ICON_PATH, 'regex.svg')
ICON_SAVE = os.path.join(ICON_PATH, 'save.svg')
ICON_SEARCH = os.path.join(ICON_PATH, 'search.svg')
ICON_SETTING = os.path.join(ICON_PATH, 'setting.svg')
ICON_SIDEBAR = os.path.join(ICON_PATH, 'sidebar.svg')
ICON_STACKONTOP = os.path.join(ICON_PATH, 'stackontop.svg')
ICON_USER = os.path.join(ICON_PATH, 'user.png')
ICON_VERTICAL_THREE_DOTS = os.path.join(ICON_PATH, 'vertical_three_dots.svg')
ICON_WORD = os.path.join(ICON_PATH, 'word.svg')
ICON_SEND = os.path.join(ICON_PATH, 'send.svg')
ICON_RECORD = os.path.join(ICON_PATH, 'record.svg')
ICON_SPEAKER = os.path.join(ICON_PATH, 'speaker.svg')
ICON_PAYPAL = os.path.join(ICON_PATH, 'paypal.png')
ICON_KOFI = os.path.join(ICON_PATH, 'kofi.png')

## CUSTOMIZE
DEFAULT_ICON_SIZE = (24, 24)
DEFAULT_USER_IMAGE_PATH = ICON_USER
DEFAULT_AI_IMAGE_PATH = ICON_OPENAI
DEFAULT_FONT_SIZE = 12
DEFAULT_FONT_FAMILY = 'Arial'

DEFAULT_BUTTON_HOVER_COLOR = '#A2D0DD'
DEFAULT_BUTTON_PRESSED_COLOR = '#B3E0FF'
DEFAULT_BUTTON_CHECKED_COLOR = '#B3E0FF'
DEFAULT_SOURCE_HIGHLIGHT_COLOR = '#CCB500'
DEFAULT_SOURCE_ERROR_COLOR = '#FF0000'
DEFAULT_FOUND_TEXT_COLOR = '#00A2E8'
DEFAULT_FOUND_TEXT_BG_COLOR = '#FFF200'

DEFAULT_LINK_COLOR = '#4F93FF'
DEFAULT_LINK_HOVER_COLOR = '#FF0000'

DEFAULT_TOAST_BACKGROUND_COLOR = '#444444'
DEFAULT_TOAST_FOREGROUND_COLOR = '#EEEEEE'

## MARKDOWN
# I am not planning to use it at the moment.
# DEFAULT_MARKDOWN_span_font = 'Courier New'
# DEFAULT_MARKDOWN_span_color = '#000'
# DEFAULT_MARKDOWN_ul_color = '#000'
# DEFAULT_MARKDOWN_h1_color = '#000'
# DEFAULT_MARKDOWN_h2_color = '#000'
# DEFAULT_MARKDOWN_h3_color = '#000'
# DEFAULT_MARKDOWN_h4_color = '#000'
# DEFAULT_MARKDOWN_h5_color = '#000'
# DEFAULT_MARKDOWN_h6_color = '#000'
# DEFAULT_MARKDOWN_a_color = '#000'

## SHORTCUT
DEFAULT_SHORTCUT_GENERAL_ACTION = 'Return'
DEFAULT_SHORTCUT_FIND_PREV = 'Ctrl+Shift+D'
DEFAULT_SHORTCUT_FIND_NEXT = 'Ctrl+D'
DEFAULT_SHORTCUT_FIND_CLOSE = 'Escape'
DEFAULT_SHORTCUT_PROMPT_BEGINNING = 'Ctrl+B'
DEFAULT_SHORTCUT_PROMPT_ENDING = 'Ctrl+E'
DEFAULT_SHORTCUT_SUPPORT_PROMPT_COMMAND = 'Ctrl+Shift+P'
DEFAULT_SHORTCUT_STACK_ON_TOP = 'Ctrl+Shift+S'
DEFAULT_SHORTCUT_SHOW_TOOLBAR = 'Ctrl+T'
DEFAULT_SHORTCUT_SHOW_SECONDARY_TOOLBAR = 'Ctrl+Shift+T'
DEFAULT_SHORTCUT_FOCUS_MODE = 'F10'
DEFAULT_SHORTCUT_FULL_SCREEN = 'F11'
DEFAULT_SHORTCUT_FIND = 'Ctrl+F'
DEFAULT_SHORTCUT_JSON_MODE = 'Ctrl+J'
DEFAULT_SHORTCUT_LEFT_SIDEBAR_WINDOW = 'Ctrl+L'
DEFAULT_SHORTCUT_RIGHT_SIDEBAR_WINDOW = 'Ctrl+R'
DEFAULT_SHORTCUT_CONTROL_PROMPT_WINDOW = 'Ctrl+Shift+C'
DEFAULT_SHORTCUT_RECORD = 'Ctrl+Shift+R'
DEFAULT_SHORTCUT_SETTING = 'Ctrl+Alt+S'
DEFAULT_SHORTCUT_SEND = 'Ctrl+Return'

## DIRECTORY PATH & FILE'S NAME
MAIN_INDEX = 'main.py'
IMAGE_DEFAULT_SAVE_DIRECTORY = 'image_result'
INI_FILE_NAME = os.path.join(get_config_directory(), 'config.yaml')
LANGUAGE_FILE_BASE_NAME = 'translations.json'
LANGUAGE_FILE = os.path.join(get_config_directory(), LANGUAGE_FILE_BASE_NAME)
LANGUAGE_FILE_SRC = os.path.join(os.path.join(EXEC_PATH, 'lang'), LANGUAGE_FILE_BASE_NAME)

# Make sure the language file exists
if not os.path.exists(LANGUAGE_FILE):
    shutil.copy(LANGUAGE_FILE_SRC, LANGUAGE_FILE)


DB_FILE_NAME = 'conv'
FILE_NAME_LENGTH = 32
QFILEDIALOG_DEFAULT_DIRECTORY = os.path.expanduser('~')

## EXTENSIONS
TEXT_FILE_EXT_LIST = ['.txt']
IMAGE_FILE_EXT_LIST = ['.png', '.jpg', '.jpeg', '.gif', '.bmp']
IMAGE_FILE_EXT_LIST_STR = 'Image File (*.png *.jpg *.jpeg *.gif *.bmp)'
TEXT_FILE_EXT_LIST_STR = 'Text File (*.txt)'
JSON_FILE_EXT_LIST_STR = 'JSON File (*.json)'
READ_FILE_EXT_LIST_STR = f'{TEXT_FILE_EXT_LIST_STR};;{IMAGE_FILE_EXT_LIST_STR}'

## PROMPT
PROMPT_BEGINNING_KEY_NAME = 'prompt_beginning'
PROMPT_JSON_KEY_NAME = 'prompt_json'
PROMPT_MAIN_KEY_NAME = 'prompt_main'
PROMPT_END_KEY_NAME = 'prompt_ending'
PROMPT_NAME_REGEX = '^[a-zA-Z_0-9]+$'
INDENT_SIZE = 4
NOTIFIER_MAX_CHAR = 100

# DB
DB_NAME_REGEX = '[a-zA-Z0-9]{1,20}'

THREAD_TABLE_NAME_OLD = 'conv_tb'
THREAD_TRIGGER_NAME_OLD = 'conv_tr'
MESSAGE_TABLE_NAME_OLD = 'conv_unit_tb'

THREAD_TABLE_NAME = 'thread_tb'
THREAD_TRIGGER_NAME = 'thread_tr'
MESSAGE_TABLE_NAME = 'message_tb'

IMAGE_TABLE_NAME = 'image_tb'

THREAD_MESSAGE_INSERTED_TR_NAME_OLD = 'conv_tb_updated_by_unit_inserted_tr'
THREAD_MESSAGE_UPDATED_TR_NAME_OLD = 'conv_tb_updated_by_unit_updated_tr'
THREAD_MESSAGE_DELETED_TR_NAME_OLD = 'conv_tb_updated_by_unit_deleted_tr'

THREAD_MESSAGE_INSERTED_TR_NAME = 'thread_message_inserted_tr'
THREAD_MESSAGE_UPDATED_TR_NAME = 'thread_message_updated_tr'
THREAD_MESSAGE_DELETED_TR_NAME = 'thread_message_deleted_tr'

THREAD_ORDERBY = 'update_dt'

PROPERTY_PROMPT_GROUP_TABLE_NAME_OLD = 'prop_prompt_grp_tb'
PROPERTY_PROMPT_UNIT_TABLE_NAME_OLD = 'prop_prompt_unit_tb'
TEMPLATE_PROMPT_GROUP_TABLE_NAME_OLD = 'template_prompt_grp_tb'
TEMPLATE_PROMPT_TABLE_NAME_OLD = 'template_prompt_tb'
PROPERTY_PROMPT_UNIT_DEFAULT_VALUE = [{'name': 'Task', 'content': ''},
                                      {'name': 'Topic', 'content': ''},
                                      {'name': 'Style', 'content': ''},
                                      {'name': 'Tone', 'content': ''},
                                      {'name': 'Audience', 'content': ''},
                                      {'name': 'Length', 'content': ''},
                                      {'name': 'Form', 'content': ''}]

PROMPT_GROUP_TABLE_NAME = 'prompt_group_tb'
PROMPT_ENTRY_TABLE_NAME = 'prompt_entry_tb'

# AI
## OPENAI
OPENAI_REQUEST_URL = 'https://api.openai.com/v1/models'

## PARAMETER - OPENAI CHAT
OPENAI_TEMPERATURE_RANGE = 0, 2
OPENAI_TEMPERATURE_STEP = 0.01

MAX_TOKENS_RANGE = 512, 128000

TOP_P_RANGE = 0, 1
TOP_P_STEP = 0.01

FREQUENCY_PENALTY_RANGE = 0, 2
FREQUENCY_PENALTY_STEP = 0.01

PRESENCE_PENALTY_RANGE = 0, 2
PRESENCE_PENALTY_STEP = 0.01

## OPENAI WHISPER (TTS, STT)
WHISPER_TTS_VOICE_TYPE = ['alloy', 'echo', 'fable', 'onyx', 'nova', 'shimmer']
WHISPER_TTS_VOICE_SPEED_RANGE = 0.25, 4.0
WHISPER_TTS_MODEL = 'tts-1'
WHISPER_TTS_DEFAULT_VOICE = 'alloy'
WHISPER_TTS_DEFAULT_SPEED = 1.0

STT_MODEL = 'whisper-1'

# Endpoint
# https://platform.openai.com/docs/models/model-endpoint-compatibility
OPENAI_ENDPOINT_DICT = {
    '/v1/chat/completions': ['gpt-4o', 'gpt-4o-mini', 'o1-preview', 'o1-mini'],
    '/v1/completions': [
        'text-davinci-003', 'text-davinci-002', 'text-curie-001', 'text-babbage-001', 'text-ada-001', 'davinci',
        'curie', 'babbage', 'ada'
    ],
    '/v1/edits': ['text-davinci-edit-001', 'code-davinci-edit-001'],
    '/v1/audio/transcriptions': ['whisper-1'],
    '/v1/audio/translations': ['whisper-1'],
    '/v1/fine-tunes': ['davinci', 'curie', 'babbage', 'ada'],
    '/v1/embeddings': ['text-embedding-ada-002', 'text-search-ada-doc-001'],
    '/vi/moderations': ['text-moderation-stable', 'text-moderation-latest']
}

# This doesn't need endpoint
DALLE_ARR = ['dall-e-2', 'dall-e-3']

DEFAULT_DATETIME_FORMAT = '%Y-%m-%d %H:%M:%S'

OPENAI_CHAT_ENDPOINT = '/v1/chat/completions'

# Other models' configuration data
DEFAULT_GEMINI_MODEL = 'gemini-1.5-flash'
LLAMA_REQUEST_URL = "https://api.llama-api.com"

# Overall API configuration data
DEFAULT_API_CONFIGS = [
    {
        "display_name": "OpenAI",
        "env_var_name": "OPENAI_API_KEY",
        "api_key": ''
    },
    {
        "display_name": "Gemini",
        "env_var_name": "GEMINI_API_KEY",
        "api_key": ''
    },
    {
        "display_name": "Claude",
        "env_var_name": "CLAUDE_API_KEY",
        "api_key": ''
    },
    {
        "display_name": "Llama",
        "env_var_name": "LLAMA_API_KEY",
        "api_key": ''
    }
]

# This has to be managed separately since some of the arguments are different with usual models
O1_MODELS = ['o1-preview', 'o1-mini']

DEFAULT_LLM = 'gpt-4o'

G4F_PROVIDER_DEFAULT = 'Auto'

G4F_USE_CHAT_HISTORY = True

# Dictionary that stores the platform and model pairs
PROVIDER_MODEL_DICT = {
    'OpenAI': ['gpt-4o', 'gpt-4o-mini']+O1_MODELS,
    'Gemini': ['gemini-1.5-flash'],
    'Claude': ['claude-3-5-sonnet-20240620'],
    'Llama': ['llama3-70b']
}

# Constants related to the number of messages LLM will store
MAXIMUM_MESSAGES_IN_PARAMETER = 40
MAXIMUM_MESSAGES_IN_PARAMETER_RANGE = 2, 1000

# PROMPT
## DEFAULT JSON FILENAME FOR PROMPT
AWESOME_CHATGPT_PROMPTS_FILENAME = 'prompt_res/awesome_chatgpt_prompts.json'
ALEX_BROGAN_PROMPT_FILENAME = 'prompt_res/alex_brogan.json'

FORM_PROMPT_GROUP_SAMPLE = json.dumps([
    {
        "name": 'Default',
        "data": PROPERTY_PROMPT_UNIT_DEFAULT_VALUE
    }
], indent=INDENT_SIZE)

SENTENCE_PROMPT_GROUP_SAMPLE = '''[
    {
        "name": "alex_brogan",
        "data": [
            {
                "name": "sample_1",
                "content": "Identify the 20% of [topic or skill] that will yield 80% of the desired results and provide a focused learning plan to master it."
            },
            {
                "name": "sample_2",
                "content": "Explain [topic or skill] in the simplest terms possible as if teaching it to a complete beginner. Identify gaps in my understanding and suggest resources to fill them."
            }
        ]
    },
    {
        "name": "awesome_chatGPT_prompts",
        "data": [
            {
                "name": "linux_terminal",
                "content": "I want you to act as a linux terminal. I will type commands and you will reply with what the terminal should show. I want you to only reply with the terminal output inside one unique code block, and nothing else. do not write explanations. do not type commands unless I instruct you to do so. when i need to tell you something in english, i will do so by putting text inside curly brackets {like this}. my first command is pwd"
            },
            {
                "name": "english_translator_and_improver",
                "content": "I want you to act as an English translator, spelling corrector and improver. I will speak to you in any language and you will detect the language, translate it and answer in the corrected and improved version of my text, in English. I want you to replace my simplified A0-level words and sentences with more beautiful and elegant, upper level English words and sentences. Keep the meaning same, but make them more literary. I want you to only reply the correction, the improvements and nothing else, do not write explanations. My first sentence is \"istanbulu cok seviyom burada olmak cok guzel\""
            },
        ]
    }
]'''

## Load the default prompt
if os.path.exists(AWESOME_CHATGPT_PROMPTS_FILENAME):
    AWESOME_CHATGPT_PROMPTS = json.load(open(AWESOME_CHATGPT_PROMPTS_FILENAME))[0]
if os.path.exists(ALEX_BROGAN_PROMPT_FILENAME):
    ALEX_BROGAN_PROMPT = json.load(open(ALEX_BROGAN_PROMPT_FILENAME))[0]

## Data for random prompt generating feature for image generation
hair_color_randomizer = ['blonde hair', 'red hair', 'blue hair', 'dark hair', 'green hair', 'white hair']
hair_style_randomizer = ['Long straight hair', 'Short pixie cut', 'Bob haircut', 'Layered haircut', 'Curly hair', 'Wavy hair', 'Ponytail', 'Messy bun', 'French braid', 'Dutch braid', 'Fishtail braid', 'High bun', 'Low bun', 'Top knot', 'Half-up half-down hairstyle', 'Braided crown', 'Side-swept bangs', 'Bangs/fringe', 'Chignon', 'Waterfall braid', 'Mohawk', 'Beach waves', 'Updo', 'French twist', 'Cornrows', 'Dreadlocks', 'Pigtails', 'Space buns', 'Sock bun']
clothes_randomizer = ['wearing denim jacket', 'wearing coat', 'wearing blazer', 'wearing blouse', 'wearing biker clothes', 'wearing distressed jeans', 'wearing hoodie']
expression_randomizer = ['Smile', 'Grin', 'Closed eyes', 'Squinting', '(laughing:1.2)', '(giggle:1.2)', '(Angry face:1.4)', 'Frowning face', 'Confused face', 'Boredom', 'Eyebrow furrow', 'Eye roll', 'Smirk']
pose_randomizer = ['sitting', 'lying', 'leaning', 'cross-legged']
camera_distance_randomizer = ['(extreme close-up:1.5)', '(medium full shot:1.5)', '(close-up:1.5)', '(establishing shot:1.5)', '(medium close-up:1.5)', '(point-of-view:1.5)', '(medium shot:1.5)', '(cowboy shot:1.5)', '(long shot:1.5)', '(upper body:1.5)', '(full shot:1.5)', '(full body:1.5)']
camera_angle_randomizer = ['(front view:1.5)', '(from below:1.5)', '(bilaterally symmetrical:1.5)', '(from behind:1.5)', '(side view:1.5)', '(wide angle view:1.5)', '(back view:1.5)', '(fisheyes view:1.5)',
                           '(from above:1.5)', '(macro view:1.5)',
                           '(overhead shot:1.5)', '(straight on:1.5)', '(top down view:1.5)', '(hero view:1.5)',
                           "(bird's eye view:1.5)", '(low view:1.5)', '(high angle:1.5)',
                           "(worm's eye view:1.5)", '(slightly above:1.5)', '(selfie:1.5)']

adding_lighting_randomizer = ['bloom', 'backlight', 'sun light', 'soft lighting', 'god rays', 'studio light', 'hard lighting', 'volumetic lighting', 'bioluminescent light']
wind_randomizer = ['(wind:0)', '(wind:0.5)', '(wind:1)', '(wind:1.5)', '(wind:2)']
acc1_randomizer = ['wearing sunglasses', 'tattoo']
acc3_randomizer = ['wearing beanie', 'wearing beret', 'wearing cap', 'wearing fedora']
attr2_randomizer = ['put on earphones', 'piercing on nose', 'put on headphones']
location_randomizer = ['In the office', 'Tokyo street', 'In the living room', 'In the music studio', 'On the beach', 'In the club']

RANDOMIZING_PROMPT_SOURCE_ARR = [
    hair_color_randomizer,
    hair_style_randomizer,
    clothes_randomizer,
    expression_randomizer,
    pose_randomizer,
    camera_distance_randomizer,
    camera_angle_randomizer,
    adding_lighting_randomizer,
    wind_randomizer,
    acc1_randomizer,
]

# DEFAULT Configuration data for the application settings
# Initialize here to avoid circular import
# ----------------------------
CONFIG_DATA = {
    'General': {
        'TAB_IDX': 0,
        'lang': 'English',
        'show_chat_list': True,
        'stream': True,
        'db': 'conv',
        'model': DEFAULT_LLM,
        'show_setting': True,
        'use_llama_index': False,
        'do_not_ask_again': False,
        'show_prompt': True,
        'system': 'You are a helpful assistant.',
        'notify_finish': True,
        'temperature': 1,
        'max_tokens': -1,
        'show_toolbar': True,
        'show_secondary_toolbar': True,
        'top_p': 1,
        'chat_column_to_show': ['id', 'name', 'insert_dt', 'update_dt'],
        'frequency_penalty': 0,
        'image_column_to_show': ['id', 'model', 'width', 'height', 'prompt', 'negative_prompt', 'n', 'quality', 'data', 'style', 'revised_prompt', 'update_dt', 'insert_dt'],
        'presence_penalty': 0,
        'json_object': False,
        'maximum_messages_in_parameter': MAXIMUM_MESSAGES_IN_PARAMETER,
        'show_as_markdown': True,
        'run_at_startup': True,
        'use_max_tokens': False,
        'background_image': '',
        'user_image': DEFAULT_USER_IMAGE_PATH,
        'ai_image': DEFAULT_AI_IMAGE_PATH,
        'font_size': DEFAULT_FONT_SIZE,
        'font_family': DEFAULT_FONT_FAMILY,
        'llama_index_directory': '',
        'apply_user_defined_styles': False,
        'focus_mode': False,
        'voice': WHISPER_TTS_DEFAULT_VOICE,
        'voice_speed': WHISPER_TTS_DEFAULT_SPEED,

        'OPENAI_API_KEY': '',
        'GEMINI_API_KEY': '',
        'CLAUDE_API_KEY': '',
        'LLAMA_API_KEY': '',

        # G4F
        'g4f_model': DEFAULT_LLM,
        'provider': G4F_PROVIDER_DEFAULT,
        'g4f_use_chat_history': G4F_USE_CHAT_HISTORY,
    },
    'DALLE': {
        'quality': 'standard',
        'show_history': True,
        'n': 1,
        'show_setting': True,
        'size': '1024x1024',
        'directory': QFILEDIALOG_DEFAULT_DIRECTORY,
        'is_save': True,
        'continue_generation': False,
        'number_of_images_to_create': 2,
        'style': 'vivid',
        'response_format': 'b64_json',
        'save_prompt_as_text': True,
        'show_prompt_on_image': False,
        'prompt_type': 1,
        'width': 1024,
        'height': 1024,
        'prompt': "Astronaut in a jungle, cold color palette, muted colors, detailed, 8k"
    },
    'REPLICATE': {
        'REPLICATE_API_TOKEN': '',
        'show_history': True,
        'model': 'stability-ai/sdxl:39ed52f2a78e934b3ba6e2a89f5b1c712de7dfea535525255b1aa35c5565e08b',
        'show_setting': True,
        'width': 768,
        'height': 768,
        'prompt': "Astronaut in a jungle, cold color palette, muted colors, detailed, 8k",
        'negative_prompt': "ugly, deformed, noisy, blurry, distorted",
        'directory': QFILEDIALOG_DEFAULT_DIRECTORY,
        'is_save': True,
        'continue_generation': False,
        'number_of_images_to_create': 2,
        'save_prompt_as_text': True,
        'show_prompt_on_image': False
    }
}

# Dynamically add the API keys to the configuration data
def update_general_config_with_api_keys(config_data, api_configs):
    for config in api_configs:
        config_data['General'][config['env_var_name']] = config['api_key']

update_general_config_with_api_keys(CONFIG_DATA, DEFAULT_API_CONFIGS)

# Update the __all__ list with the PEP8 standard dunder names
__all__ = ['__version__',
           '__author__']

# Update the __all__ list with the constants
__all__.extend([name for name, value in globals().items() if name.isupper()])


from PySide6.QtCore import Qt
from PySide6.QtGui import QFont
from PySide6.QtWidgets import QDialog, QPushButton, QHBoxLayout, QVBoxLayout, QWidget, QFormLayout, QSplitter, \
    QSizePolicy

from pyqt_openai import IMAGE_FILE_EXT_LIST_STR, DEFAULT_ICON_SIZE
from pyqt_openai.fontWidget import FontWidget
from pyqt_openai.lang.translations import LangClass
from pyqt_openai.models import CustomizeParamsContainer
from pyqt_openai.util.script import getSeparator
from pyqt_openai.widgets.circleProfileImage import RoundedImage
from pyqt_openai.widgets.findPathWidget import FindPathWidget
from pyqt_openai.widgets.normalImageView import NormalImageView


class CustomizeDialog(QDialog):
    def __init__(self, args: CustomizeParamsContainer, parent=None):
        super().__init__(parent)
        self.__initVal(args)
        self.__initUi()

    def __initVal(self, args):
        self.__background_image = args.background_image
        self.__user_image = args.user_image
        self.__ai_image = args.ai_image
        self.__font_size = args.font_size
        self.__font_family = args.font_family

    def __initUi(self):
        self.setWindowTitle(LangClass.TRANSLATIONS['Customize'])
        self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowCloseButtonHint)

        self.__homePageGraphicsView = NormalImageView()
        self.__homePageGraphicsView.setFilename(self.__background_image)

        self.__userImage = RoundedImage()
        self.__userImage.setMaximumSize(*DEFAULT_ICON_SIZE)
        self.__userImage.setImage(self.__user_image)
        self.__AIImage = RoundedImage()
        self.__AIImage.setImage(self.__ai_image)
        self.__AIImage.setMaximumSize(*DEFAULT_ICON_SIZE)

        self.__findPathWidget1 = FindPathWidget()
        self.__findPathWidget1.getLineEdit().setText(self.__background_image)
        self.__findPathWidget1.added.connect(self.__homePageGraphicsView.setFilename)
        self.__findPathWidget1.setExtOfFiles(IMAGE_FILE_EXT_LIST_STR)

        self.__findPathWidget2 = FindPathWidget()
        self.__findPathWidget2.getLineEdit().setText(self.__user_image)
        self.__findPathWidget2.added.connect(self.__userImage.setImage)
        self.__findPathWidget2.setExtOfFiles(IMAGE_FILE_EXT_LIST_STR)

        self.__findPathWidget3 = FindPathWidget()
        self.__findPathWidget3.getLineEdit().setText(self.__ai_image)
        self.__findPathWidget3.added.connect(self.__AIImage.setImage)
        self.__findPathWidget3.setExtOfFiles(IMAGE_FILE_EXT_LIST_STR)

        lay1 = QVBoxLayout()
        lay1.setContentsMargins(0, 0, 0, 0)
        lay1.addWidget(self.__homePageGraphicsView)
        lay1.addWidget(self.__findPathWidget1)
        homePageWidget = QWidget()
        homePageWidget.setLayout(lay1)

        lay2 = QHBoxLayout()
        lay2.setContentsMargins(0, 0, 0, 0)
        lay2.addWidget(self.__userImage)
        lay2.addWidget(self.__findPathWidget2)
        userWidget = QWidget()
        userWidget.setLayout(lay2)

        lay3 = QHBoxLayout()
        lay3.setContentsMargins(0, 0, 0, 0)
        lay3.addWidget(self.__AIImage)
        lay3.addWidget(self.__findPathWidget3)
        aiWidget = QWidget()
        aiWidget.setLayout(lay3)

        lay = QFormLayout()
        lay.addRow(LangClass.TRANSLATIONS['Home Image'], homePageWidget)
        lay.addRow(LangClass.TRANSLATIONS['User Image'], userWidget)
        lay.addRow(LangClass.TRANSLATIONS['AI Image'], aiWidget)

        leftWidget = QWidget()
        leftWidget.setLayout(lay)

        self.__fontWidget = FontWidget(QFont(self.__font_family, self.__font_size))

        self.__splitter = QSplitter()
        self.__splitter.addWidget(leftWidget)
        self.__splitter.addWidget(self.__fontWidget)
        self.__splitter.setHandleWidth(1)
        self.__splitter.setChildrenCollapsible(False)
        self.__splitter.setSizes([500, 500])
        self.__splitter.setStyleSheet("QSplitterHandle {background-color: lightgray;}")
        self.__splitter.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.MinimumExpanding)

        sep = getSeparator('horizontal')

        self.__okBtn = QPushButton(LangClass.TRANSLATIONS['OK'])
        self.__okBtn.clicked.connect(self.accept)

        cancelBtn = QPushButton(LangClass.TRANSLATIONS['Cancel'])
        cancelBtn.clicked.connect(self.close)

        lay = QHBoxLayout()
        lay.addWidget(self.__okBtn)
        lay.addWidget(cancelBtn)
        lay.setAlignment(Qt.AlignmentFlag.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        okCancelWidget = QWidget()
        okCancelWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(self.__splitter)
        lay.addWidget(sep)
        lay.addWidget(okCancelWidget)

        self.setLayout(lay)

    def getParam(self):
        return CustomizeParamsContainer(
            background_image=self.__findPathWidget1.getFileName(),
            user_image=self.__findPathWidget2.getFileName(),
            ai_image=self.__findPathWidget3.getFileName(),
            font_size=self.__fontWidget.getFont().pointSize(),
            font_family=self.__fontWidget.getFont().family()
        )


import subprocess
import sys

import requests
from PySide6.QtCore import Qt
from PySide6.QtWidgets import QDialog, QVBoxLayout, QLabel, QTextBrowser, QDialogButtonBox, QMessageBox

from pyqt_openai import __version__, OWNER, PACKAGE_NAME, UPDATE_DIR, CURRENT_FILENAME, UPDATER_PATH, is_frozen


class UpdateSoftwareDialog(QDialog):
    def __init__(self, owner, repo, recent_version, parent=None):
        super().__init__(parent)
        self.__initVal(owner, repo, recent_version)
        self.__initUi()

    def __initVal(self, owner, repo, recent_version):
        self.__owner = owner
        self.__repo = repo
        self.__recent_version = f'v{recent_version}'

    def __initUi(self):
        # TODO LANGUAGE
        self.setWindowTitle("Update Software")
        self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowCloseButtonHint)
        self.setModal(True)
        lay = QVBoxLayout()

        self.setLayout(lay)

        self.__lbl = QLabel("A new version of the software is available.")
        lay.addWidget(self.__lbl)

        self.releaseNoteBrowser = QTextBrowser()
        self.releaseNoteBrowser.setOpenExternalLinks(True)

        update_url = f"https://github.com/{self.__owner}/{self.__repo}/releases/download/{self.__recent_version}/VividNode.zip"

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(lambda: run_updater(update_url))
        buttonBox.rejected.connect(self.reject)

        askLbl = QLabel("Do you want to update?")

        lay.addWidget(self.releaseNoteBrowser)
        lay.addWidget(askLbl)
        lay.addWidget(buttonBox)


def check_for_updates(current_version, owner, repo):
    try:
        url = f"https://api.github.com/repos/{owner}/{repo}/releases"

        response = requests.get(url)
        releases = response.json()

        update_available = False
        release_notes_html = "<ul>"
        recent_version = current_version
        for release in releases:
            release_version = release['tag_name'].lstrip('v')
            if release_version > current_version:
                recent_version = release_version if recent_version < release_version else recent_version
                update_available = True
                release_notes_html += f'<li><a href="{release["html_url"]}" target="_blank">{release["tag_name"]}</a></li>'
        release_notes_html += "</ul>"

        if update_available:
            return {'release_notes': release_notes_html, 'recent_version': recent_version}
        else:
            return None

    except Exception as e:
        QMessageBox.critical(None, "Error", f"Error fetching release notes: {str(e)}")
        return None


def check_for_updates_and_show_dialog(current_version, owner, repo):
    result_dict = check_for_updates(current_version, owner, repo)
    if result_dict:
        release_notes = result_dict['release_notes']
        recent_version = result_dict['recent_version']
        if release_notes:
            # If updates are available, show the update dialog
            update_dialog = UpdateSoftwareDialog(owner, repo, recent_version)
            update_dialog.releaseNoteBrowser.setHtml(release_notes)
            update_dialog.exec()


def update_software():
    # Replace with actual values
    current_version = __version__
    owner = OWNER
    repo = PACKAGE_NAME

    if not is_frozen():
        return

    # Check for updates and show dialog if available
    check_for_updates_and_show_dialog(current_version, owner, repo)


def run_updater(update_url):
    subprocess.Popen([UPDATER_PATH, update_url, UPDATE_DIR, CURRENT_FILENAME], shell=True)
    sys.exit(0)



import webbrowser

from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QIcon
from PySide6.QtWidgets import QMainWindow, QToolBar, QHBoxLayout, QDialog, QWidgetAction, QSpinBox, \
    QWidget, QApplication, \
    QSizePolicy, QStackedWidget, QMenu, QSystemTrayIcon, \
    QMessageBox

from pyqt_openai import DEFAULT_SHORTCUT_FULL_SCREEN, \
    APP_INITIAL_WINDOW_SIZE, DEFAULT_APP_NAME, DEFAULT_APP_ICON, ICON_STACKONTOP, ICON_CUSTOMIZE, ICON_FULLSCREEN, \
    ICON_CLOSE, \
    DEFAULT_SHORTCUT_SETTING, TRANSPARENT_RANGE, TRANSPARENT_INIT_VAL, ICON_GITHUB, ICON_DISCORD, PAYPAL_URL, KOFI_URL, \
    DISCORD_URL, GITHUB_URL, DEFAULT_SHORTCUT_FOCUS_MODE, ICON_FOCUS_MODE, ICON_SETTING, DEFAULT_SHORTCUT_SHOW_TOOLBAR, \
    DEFAULT_SHORTCUT_SHOW_SECONDARY_TOOLBAR, DEFAULT_SHORTCUT_STACK_ON_TOP, ICON_PAYPAL, ICON_KOFI
from pyqt_openai.aboutDialog import AboutDialog
from pyqt_openai.chat_widget.chatMainWidget import ChatMainWidget
from pyqt_openai.config_loader import CONFIG_MANAGER
from pyqt_openai.customizeDialog import CustomizeDialog
from pyqt_openai.dalle_widget.dalleMainWidget import DallEMainWidget
from pyqt_openai.doNotAskAgainDialog import DoNotAskAgainDialog
from pyqt_openai.lang.translations import LangClass
from pyqt_openai.models import SettingsParamsContainer, CustomizeParamsContainer
from pyqt_openai.replicate_widget.replicateMainWidget import ReplicateMainWidget
from pyqt_openai.settings_dialog.settingsDialog import SettingsDialog
from pyqt_openai.shortcutDialog import ShortcutDialog
from pyqt_openai.updateSoftwareDialog import update_software
from pyqt_openai.util.script import restart_app, show_message_box_after_change_to_restart, set_auto_start_windows, \
    set_api_key, init_llama
from pyqt_openai.widgets.navWidget import NavBar


class MainWindow(QMainWindow):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.__initVal()
        self.__initUi()

    def __initVal(self):
        self.__settingsParamContainer = SettingsParamsContainer()
        self.__customizeParamsContainer = CustomizeParamsContainer()

        self.__initContainer(self.__settingsParamContainer)
        self.__initContainer(self.__customizeParamsContainer)

    def __initUi(self):
        self.setWindowTitle(DEFAULT_APP_NAME)

        self.__chatMainWidget = ChatMainWidget(self)
        self.__dallEWidget = DallEMainWidget(self)
        self.__replicateWidget = ReplicateMainWidget(self)

        self.__mainWidget = QStackedWidget()
        self.__mainWidget.addWidget(self.__chatMainWidget)
        self.__mainWidget.addWidget(self.__dallEWidget)
        self.__mainWidget.addWidget(self.__replicateWidget)

        self.__setActions()
        self.__setMenuBar()
        self.__setTrayMenu()
        self.__setToolBar()

        self.__loadApiKeyInConf()

        self.setCentralWidget(self.__mainWidget)
        self.resize(*APP_INITIAL_WINDOW_SIZE)

        self.__refreshColumns()
        self.__chatMainWidget.refreshCustomizedInformation(self.__customizeParamsContainer)

    def __setActions(self):
        self.__langAction = QAction()

        # menu action
        self.__exitAction = QAction(LangClass.TRANSLATIONS['Exit'], self)
        self.__exitAction.triggered.connect(self.__beforeClose)

        self.__stackAction = QAction(LangClass.TRANSLATIONS['Stack on Top'], self)
        self.__stackAction.setShortcut(DEFAULT_SHORTCUT_STACK_ON_TOP)
        self.__stackAction.setIcon(QIcon(ICON_STACKONTOP))
        self.__stackAction.setCheckable(True)
        self.__stackAction.toggled.connect(self.__stackToggle)

        self.__showToolBarAction = QAction(LangClass.TRANSLATIONS['Show Toolbar'], self)
        self.__showToolBarAction.setShortcut(DEFAULT_SHORTCUT_SHOW_TOOLBAR)
        self.__showToolBarAction.setCheckable(True)
        self.__showToolBarAction.setChecked(CONFIG_MANAGER.get_general_property('show_toolbar'))
        self.__showToolBarAction.toggled.connect(self.__toggleToolbar)

        self.__showSecondaryToolBarAction = QAction(LangClass.TRANSLATIONS['Show Secondary Toolbar'], self)
        self.__showSecondaryToolBarAction.setShortcut(DEFAULT_SHORTCUT_SHOW_SECONDARY_TOOLBAR)
        self.__showSecondaryToolBarAction.setCheckable(True)
        self.__showSecondaryToolBarAction.setChecked(CONFIG_MANAGER.get_general_property('show_secondary_toolbar'))
        self.__showSecondaryToolBarAction.toggled.connect(self.__toggleSecondaryToolBar)

        self.__focusModeAction = QAction(LangClass.TRANSLATIONS['Focus Mode'], self)
        self.__focusModeAction.setShortcut(DEFAULT_SHORTCUT_FOCUS_MODE)
        self.__focusModeAction.setIcon(QIcon(ICON_FOCUS_MODE))
        self.__focusModeAction.setCheckable(True)
        self.__focusModeAction.setChecked(CONFIG_MANAGER.get_general_property('focus_mode'))
        self.__focusModeAction.triggered.connect(self.__activateFocusMode)

        self.__fullScreenAction = QAction(LangClass.TRANSLATIONS['Full Screen'], self)
        self.__fullScreenAction.setShortcut(DEFAULT_SHORTCUT_FULL_SCREEN)
        self.__fullScreenAction.setIcon(QIcon(ICON_FULLSCREEN))
        self.__fullScreenAction.setCheckable(True)
        self.__fullScreenAction.setChecked(False)
        self.__fullScreenAction.triggered.connect(self.__fullScreenToggle)

        self.__aboutAction = QAction(LangClass.TRANSLATIONS['About...'], self)
        self.__aboutAction.triggered.connect(self.__showAboutDialog)

        # TODO LANGAUGE
        self.__checkUpdateAction = QAction(LangClass.TRANSLATIONS['Check for Updates...'], self)
        self.__checkUpdateAction.triggered.connect(self.__checkUpdate)

        self.__viewShortcutsAction = QAction(LangClass.TRANSLATIONS['View Shortcuts'], self)
        self.__viewShortcutsAction.triggered.connect(self.__showShortcutsDialog)

        self.__githubAction = QAction('Github', self)
        self.__githubAction.setIcon(QIcon(ICON_GITHUB))
        self.__githubAction.triggered.connect(lambda: webbrowser.open(GITHUB_URL))

        self.__discordAction = QAction('Discord', self)
        self.__discordAction.setIcon(QIcon(ICON_DISCORD))
        self.__discordAction.triggered.connect(lambda: webbrowser.open(DISCORD_URL))

        self.__paypalAction = QAction('Paypal', self)
        self.__paypalAction.setIcon(QIcon(ICON_PAYPAL))
        self.__paypalAction.triggered.connect(lambda: webbrowser.open(PAYPAL_URL))

        self.__kofiAction = QAction('Ko-fi ❤', self)
        self.__kofiAction.setIcon(QIcon(ICON_KOFI))
        self.__kofiAction.triggered.connect(lambda: webbrowser.open(KOFI_URL))

        self.__navBar = NavBar()
        self.__navBar.add(LangClass.TRANSLATIONS['Chat'])
        self.__navBar.add('DALL-E')
        self.__navBar.add('Replicate')
        self.__navBar.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.Preferred)
        self.__navBar.itemClicked.connect(self.__aiTypeChanged)

        # Chat is the default widget
        self.__navBar.setActiveButton(0)

        # toolbar action
        self.__chooseAiAction = QWidgetAction(self)
        self.__chooseAiAction.setDefaultWidget(self.__navBar)

        self.__customizeAction = QAction(self)
        self.__customizeAction.setText(LangClass.TRANSLATIONS['Customize'])
        self.__customizeAction.setIcon(QIcon(ICON_CUSTOMIZE))
        self.__customizeAction.triggered.connect(self.__executeCustomizeDialog)

        self.__transparentAction = QWidgetAction(self)
        self.__transparentSpinBox = QSpinBox()
        self.__transparentSpinBox.setRange(*TRANSPARENT_RANGE)
        self.__transparentSpinBox.setValue(TRANSPARENT_INIT_VAL)
        self.__transparentSpinBox.valueChanged.connect(self.__setTransparency)
        self.__transparentSpinBox.setToolTip(LangClass.TRANSLATIONS['Set Transparency of Window'])
        self.__transparentSpinBox.setMinimumWidth(100)

        lay = QHBoxLayout()
        lay.addWidget(self.__transparentSpinBox)

        transparencyActionWidget = QWidget(self)
        transparencyActionWidget.setLayout(lay)
        self.__transparentAction.setDefaultWidget(transparencyActionWidget)

        self.__settingsAction = QAction(self)
        self.__settingsAction.setText(LangClass.TRANSLATIONS['Settings'])
        self.__settingsAction.setIcon(QIcon(ICON_SETTING))
        self.__settingsAction.setShortcut(DEFAULT_SHORTCUT_SETTING)
        self.__settingsAction.triggered.connect(self.__showSettingsDialog)

    def __fullScreenToggle(self, f):
        if f:
            self.showFullScreen()
        else:
            self.showNormal()

    def __toggleToolbar(self, f):
        self.__toolbar.setVisible(f)
        CONFIG_MANAGER.set_general_property('show_toolbar', f)
        self.__showToolBarAction.setChecked(f)

    def __activateFocusMode(self, f):
        f = not f
        # Toggle GUI
        for i in range(self.__mainWidget.count()):
            currentWidget = self.__mainWidget.widget(i)
            currentWidget.showSecondaryToolBar(f)
            currentWidget.toggleButtons(f)
        self.__toggleToolbar(f)
        self.__toggleSecondaryToolBar(f)

        # Toggle container
        self.__settingsParamContainer.show_toolbar = f
        self.__settingsParamContainer.show_secondary_toolbar = f
        CONFIG_MANAGER.set_general_property('focus_mode', not f)

    def __setMenuBar(self):
        menubar = self.menuBar()

        fileMenu = QMenu(LangClass.TRANSLATIONS['File'], self)
        fileMenu.addAction(self.__settingsAction)
        fileMenu.addAction(self.__exitAction)

        viewMenu = QMenu(LangClass.TRANSLATIONS['View'], self)
        viewMenu.addAction(self.__focusModeAction)
        viewMenu.addAction(self.__fullScreenAction)
        viewMenu.addAction(self.__stackAction)
        viewMenu.addAction(self.__showToolBarAction)
        viewMenu.addAction(self.__showSecondaryToolBarAction)

        helpMenu = QMenu(LangClass.TRANSLATIONS['Help'], self)
        helpMenu.addAction(self.__aboutAction)
        helpMenu.addAction(self.__checkUpdateAction)
        helpMenu.addAction(self.__viewShortcutsAction)
        helpMenu.addAction(self.__githubAction)
        helpMenu.addAction(self.__discordAction)

        donateMenu = QMenu(LangClass.TRANSLATIONS['Donate'], self)
        donateMenu.addAction(self.__paypalAction)
        donateMenu.addAction(self.__kofiAction)

        menubar.addMenu(fileMenu)
        menubar.addMenu(viewMenu)
        menubar.addMenu(helpMenu)
        menubar.addMenu(donateMenu)

    def __setTrayMenu(self):
        # background app
        menu = QMenu()
        app = QApplication.instance()

        action = QAction("Quit", self)
        action.setIcon(QIcon(ICON_CLOSE))

        action.triggered.connect(app.quit)

        menu.addAction(action)

        tray_icon = QSystemTrayIcon(app)
        tray_icon.setIcon(QIcon(DEFAULT_APP_ICON))
        tray_icon.activated.connect(self.__activated)

        tray_icon.setContextMenu(menu)

        tray_icon.show()

    def __activated(self, reason):
        if reason == QSystemTrayIcon.ActivationReason.DoubleClick:
            self.show()

    def __setToolBar(self):
        self.__toolbar = QToolBar()
        self.__toolbar.addAction(self.__chooseAiAction)
        self.__toolbar.addAction(self.__settingsAction)
        self.__toolbar.addAction(self.__customizeAction)
        self.__toolbar.addAction(self.__githubAction)
        self.__toolbar.addAction(self.__discordAction)
        self.__toolbar.addAction(self.__paypalAction)
        self.__toolbar.addAction(self.__kofiAction)
        self.__toolbar.addAction(self.__transparentAction)
        self.__toolbar.setMovable(False)

        self.addToolBar(self.__toolbar)

        # QToolbar's layout can't be set spacing with lay.setSpacing so i've just did this instead
        self.__toolbar.setStyleSheet('QToolBar { spacing: 2px; }')

        self.__toggleToolbar(self.__settingsParamContainer.show_toolbar)
        for i in range(self.__mainWidget.count()):
            currentWidget = self.__mainWidget.widget(i)
            currentWidget.showSecondaryToolBar(self.__settingsParamContainer.show_secondary_toolbar)

    def __loadApiKeyInConf(self):
        set_api_key('OPENAI_API_KEY', CONFIG_MANAGER.get_general_property('OPENAI_API_KEY'))
        set_api_key('GEMINI_API_KEY', CONFIG_MANAGER.get_general_property('GEMINI_API_KEY'))
        set_api_key('CLAUDE_API_KEY', CONFIG_MANAGER.get_general_property('CLAUDE_API_KEY'))
        set_api_key('LLAMA_API_KEY', CONFIG_MANAGER.get_general_property('LLAMA_API_KEY'))
        set_api_key('REPLICATE_API_TOKEN', CONFIG_MANAGER.get_replicate_property('REPLICATE_API_TOKEN'))

        # Set llama index directory if it exists
        init_llama()

    def __showAboutDialog(self):
        aboutDialog = AboutDialog(self)
        aboutDialog.exec()

    def __checkUpdate(self):
        update_software()

    def __showShortcutsDialog(self):
        shortcutListWidget = ShortcutDialog(self)
        shortcutListWidget.exec()

    def __stackToggle(self, f):
        if f:
            self.setWindowFlags(self.windowFlags() | Qt.WindowType.WindowStaysOnTopHint)
        else:
            # Qt.WindowType.WindowCloseButtonHint is added to prevent the close button get deactivated
            self.setWindowFlags(self.windowFlags() & ~Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.WindowCloseButtonHint)
        self.show()

    def __setTransparency(self, v):
        self.setWindowOpacity(v / 100)

    def __toggleSecondaryToolBar(self, f):
        self.__showSecondaryToolBarAction.setChecked(f)
        self.__mainWidget.currentWidget().showSecondaryToolBar(f)
        self.__settingsParamContainer.show_secondary_toolbar = f

    def __executeCustomizeDialog(self):
        dialog = CustomizeDialog(self.__customizeParamsContainer, parent=self)
        reply = dialog.exec()
        if reply == QDialog.DialogCode.Accepted:
            container = dialog.getParam()
            self.__customizeParamsContainer = container
            self.__refreshContainer(container)
            self.__chatMainWidget.refreshCustomizedInformation(container)

    def __aiTypeChanged(self, i):
        self.__mainWidget.setCurrentIndex(i)
        self.__navBar.setActiveButton(i)
        widget = self.__mainWidget.currentWidget()
        widget.showSecondaryToolBar(self.__settingsParamContainer.show_secondary_toolbar)

    def __initContainer(self, container):
        """
        Initialize the container with the values in the settings file
        """
        for k, v in container.get_items():
            setattr(container, k, CONFIG_MANAGER.get_general_property(k))
        if isinstance(container, SettingsParamsContainer):
            self.__lang = LangClass.lang_changed(container.lang)
            set_auto_start_windows(container.run_at_startup)

    def __refreshContainer(self, container):
        if isinstance(container, SettingsParamsContainer):
            prev_db = CONFIG_MANAGER.get_general_property('db')
            prev_show_toolbar = CONFIG_MANAGER.get_general_property('show_toolbar')
            prev_show_secondary_toolbar = CONFIG_MANAGER.get_general_property('show_secondary_toolbar')
            prev_show_as_markdown = CONFIG_MANAGER.get_general_property('show_as_markdown')
            prev_run_at_startup = CONFIG_MANAGER.get_general_property('run_at_startup')

            for k, v in container.get_items():
                CONFIG_MANAGER.set_general_property(k, v)

            # If db name is changed
            if container.db != prev_db:
                QMessageBox.information(self, LangClass.TRANSLATIONS['Info'], LangClass.TRANSLATIONS["The name of the reference target database has been changed. The changes will take effect after a restart."])
            # If show_toolbar is changed
            if container.show_toolbar != prev_show_toolbar:
                self.__toggleToolbar(container.show_toolbar)
            if container.run_at_startup != prev_run_at_startup:
                set_auto_start_windows(container.run_at_startup)
            # If show_secondary_toolbar is changed
            if container.show_secondary_toolbar != prev_show_secondary_toolbar:
                for i in range(self.__mainWidget.count()):
                    currentWidget = self.__mainWidget.widget(i)
                    currentWidget.showSecondaryToolBar(container.show_secondary_toolbar)
            # If properties that require a restart are changed
            if container.lang != self.__lang or container.show_as_markdown != prev_show_as_markdown:
                change_list = []
                if container.lang != self.__lang:
                    change_list.append(LangClass.TRANSLATIONS["Language"])
                if container.show_as_markdown != prev_show_as_markdown:
                    change_list.append(LangClass.TRANSLATIONS["Show as Markdown"])
                result = show_message_box_after_change_to_restart(change_list)
                if result == QMessageBox.StandardButton.Yes:
                    restart_app()

        elif isinstance(container, CustomizeParamsContainer):
            prev_font_family = CONFIG_MANAGER.get_general_property('font_family')
            prev_font_size = CONFIG_MANAGER.get_general_property('font_size')

            for k, v in container.get_items():
                CONFIG_MANAGER.set_general_property(k, v)

            if container.font_family != prev_font_family or container.font_size != prev_font_size:
                change_list = [
                    LangClass.TRANSLATIONS["Font Change"],
                ]
                result = show_message_box_after_change_to_restart(change_list)
                if result == QMessageBox.StandardButton.Yes:
                    restart_app()

    def __refreshColumns(self):
        self.__chatMainWidget.setColumns(self.__settingsParamContainer.chat_column_to_show)
        image_column_to_show = self.__settingsParamContainer.image_column_to_show
        if image_column_to_show.__contains__('data'):
            image_column_to_show.remove('data')
        self.__dallEWidget.setColumns(self.__settingsParamContainer.image_column_to_show)
        self.__replicateWidget.setColumns(self.__settingsParamContainer.image_column_to_show)

    def __showSettingsDialog(self):
        dialog = SettingsDialog(parent=self)
        reply = dialog.exec()
        if reply == QDialog.DialogCode.Accepted:
            container = dialog.getParam()
            self.__settingsParamContainer = container
            self.__refreshContainer(container)
            self.__refreshColumns()

    def __doNotAskAgainChanged(self, value):
        self.__settingsParamContainer.do_not_ask_again = value
        self.__refreshContainer(self.__settingsParamContainer)

    def __beforeClose(self):
        if self.__settingsParamContainer.do_not_ask_again:
            app = QApplication.instance()
            app.quit()
        else:
            # Show a message box to confirm the exit or cancel or running in the background
            dialog = DoNotAskAgainDialog(self.__settingsParamContainer.do_not_ask_again, parent=self)
            dialog.doNotAskAgainChanged.connect(self.__doNotAskAgainChanged)
            reply = dialog.exec()
            if dialog.isCancel():
                return True
            else:
                if reply == QDialog.DialogCode.Accepted:
                    app = QApplication.instance()
                    app.quit()
                elif reply == QDialog.DialogCode.Rejected:
                    self.close()

    def closeEvent(self, event):
        f = self.__beforeClose()
        if f:
            event.ignore()
        else:
            return super().closeEvent(event)


import json
import os
import sqlite3
from datetime import datetime
from typing import List

from pyqt_openai import THREAD_TABLE_NAME, THREAD_TRIGGER_NAME, \
    MESSAGE_TABLE_NAME, THREAD_MESSAGE_INSERTED_TR_NAME, \
    THREAD_MESSAGE_UPDATED_TR_NAME, THREAD_MESSAGE_DELETED_TR_NAME, IMAGE_TABLE_NAME, \
    PROMPT_GROUP_TABLE_NAME, PROMPT_ENTRY_TABLE_NAME, get_config_directory, DEFAULT_DATETIME_FORMAT
from pyqt_openai.config_loader import CONFIG_MANAGER
from pyqt_openai.models import ImagePromptContainer, ChatMessageContainer, PromptEntryContainer, PromptGroupContainer


def get_db_filename():
    """
    Get the database file's name from the settings.
    """
    db_filename = CONFIG_MANAGER.get_general_property('db') + '.db'
    config_dir = get_config_directory()
    db_path = os.path.join(config_dir, db_filename)
    return db_path


class SqliteDatabase:
    """
    Functions which only meant to be used frequently are defined.
    If there is no functions you want to use, use ``getCursor`` instead.
    """
    def __init__(self, db_filename=get_db_filename()):
        super().__init__()
        self.__initVal(db_filename)
        self.__initDb()

    def __initVal(self, db_filename):
        # DB file name
        self.__db_filename = db_filename or get_db_filename()

    def __initDb(self):
        try:
            # Connect to the database (create a new file if it doesn't exist)
            self.__conn = sqlite3.connect(self.__db_filename)
            self.__conn.row_factory = sqlite3.Row
            self.__conn.execute('PRAGMA foreign_keys = ON;')
            self.__conn.commit()

            # create cursor
            self.__c = self.__conn.cursor()

            # create conversation tables
            self.__createThread()

            # create prompt tables
            self.__createPromptGroup()

            # create image tables
            self.__createImage()
        except sqlite3.Error as e:
            print(f"An error occurred while connecting to the database: {e}")
            raise

    def __createPromptGroup(self):
        try:
            self.__c.execute(
                f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{PROMPT_GROUP_TABLE_NAME}'")
            if self.__c.fetchone()[0] == 1:
                pass
            else:
                self.__c.execute(f'''CREATE TABLE {PROMPT_GROUP_TABLE_NAME}
                                     (id INTEGER PRIMARY KEY,
                                      name VARCHAR(255) UNIQUE,
                                      prompt_type VARCHAR(255),
                                      update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                      insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')
                # Create prompt entry
                self.__createPromptEntry()

                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def insertPromptGroup(self, name, prompt_type):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {PROMPT_GROUP_TABLE_NAME} (name, prompt_type) VALUES (?, ?)', (name,
                                                                                                          prompt_type))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectPromptGroup(self, prompt_type=None):
        try:
            query = f'SELECT * FROM {PROMPT_GROUP_TABLE_NAME}'
            if prompt_type == 'form':
                query += f' WHERE prompt_type="form"'
            elif prompt_type == 'sentence':
                query += f' WHERE prompt_type="sentence"'
            self.__c.execute(query)
            return [PromptGroupContainer(**elem) for elem in self.__c.fetchall()]
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectCertainPromptGroup(self, id=None, name=None):
        """
        Select specific prompt group by id or name
        """
        try:
            query = f'SELECT * FROM {PROMPT_GROUP_TABLE_NAME}'
            if id or name:
                query += ' WHERE'
                if id:
                    query += f' id={id}'
                    if name:
                        query += ' AND'
                if name:
                    query += f' name="{name}"'
            result = self.__c.execute(query).fetchone()
            if result:
                return PromptGroupContainer(**result)
            else:
                return None
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updatePromptGroup(self, id, name):
        try:
            self.__c.execute(f'UPDATE {PROMPT_GROUP_TABLE_NAME} SET name=? WHERE id={id}', (name,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deletePromptGroup(self, id=None):
        try:
            query = f'DELETE FROM {PROMPT_GROUP_TABLE_NAME}'
            if id:
                query += f' WHERE id = {id}'
            self.__c.execute(query)
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createPromptEntry(self):
        try:
            self.__c.execute(
                f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{PROMPT_ENTRY_TABLE_NAME}'")
            if self.__c.fetchone()[0] == 1:
                # Let it pass if the table already exists
                pass
            else:
                self.__c.execute(f'''CREATE TABLE {PROMPT_ENTRY_TABLE_NAME} (
                                    id INTEGER PRIMARY KEY,
                                    group_id INTEGER NOT NULL,
                                    name VARCHAR(255) NOT NULL,
                                    content TEXT NOT NULL,
                                    insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                    update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                                    FOREIGN KEY (group_id) REFERENCES {PROMPT_GROUP_TABLE_NAME}(id)
                                    ON DELETE CASCADE)
                ''')
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertPromptEntry(self, group_id, name, content=''):
        try:
            # Insert a row into the table
            self.__c.execute(f'INSERT INTO {PROMPT_ENTRY_TABLE_NAME} (group_id, name, content) VALUES (?, ?, ?)',
                             (group_id, name, content))
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectPromptEntry(self, group_id, id=None, name=None) -> List[PromptEntryContainer]:
        try:
            query = f'SELECT * FROM {PROMPT_ENTRY_TABLE_NAME} WHERE group_id={group_id}'
            if id:
                query += f' AND id={id}'
            if name:
                query += f' AND name="{name}"'
            return [PromptEntryContainer(**elem) for elem in self.__c.execute(query).fetchall()]
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updatePromptEntry(self, id, name, content):
        try:
            self.__c.execute(f'UPDATE {PROMPT_ENTRY_TABLE_NAME} SET name=?, content=? WHERE id={id}', (name, content))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deletePromptEntry(self, group_id, id=None):
        try:
            query = f'DELETE FROM {PROMPT_ENTRY_TABLE_NAME} WHERE group_id={group_id}'
            if id:
                query += f' AND id={id}'
            self.__c.execute(query)
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createThread(self):
        try:
            # Create thread table if not exists
            thread_tb_exists = self.__c.execute(
                    f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{THREAD_TABLE_NAME}'").fetchone()[
                                            0] == 1
            if thread_tb_exists:
                pass
            else:
                # If user uses app for the first time, create a table
                # Create a table with update_dt and insert_dt columns
                self.__c.execute(f'''CREATE TABLE {THREAD_TABLE_NAME}
                             (id INTEGER PRIMARY KEY,
                              name TEXT,
                              update_dt DATETIME,
                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')

            # Create message table
            self.__createMessage()

            # Create trigger if not exists
            thread_trigger_exists = self.__c.execute(
                f"SELECT count(*) FROM sqlite_master WHERE type='trigger' AND name='{THREAD_TRIGGER_NAME}'").fetchone()[
                                        0] == 1
            if thread_trigger_exists:
                pass
            else:
                # Create a trigger to update the update_dt column with the current timestamp
                self.__c.execute(f'''CREATE TRIGGER {THREAD_TRIGGER_NAME}
                             AFTER UPDATE ON {THREAD_TABLE_NAME}
                             FOR EACH ROW
                             BEGIN
                               UPDATE {THREAD_TABLE_NAME}
                               SET update_dt=CURRENT_TIMESTAMP
                               WHERE id=OLD.id;
                             END;''')
            # Commit the transaction
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectAllThread(self, id_arr=None):
        """
        Select all thread
        id_arr: list of thread id
        """
        try:
            query = f'SELECT * FROM {THREAD_TABLE_NAME}'
            if id_arr:
                query += f' WHERE id IN ({",".join(map(str, id_arr))})'
            self.__c.execute(query)
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectThread(self, id):
        """
        Select specific thread
        """
        try:
            self.__c.execute(f'SELECT * FROM {THREAD_TABLE_NAME} WHERE id={id}')
            return self.__c.fetchone()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def insertThread(self, name, insert_dt=None, update_dt=None):
        try:
            query = f'INSERT INTO {THREAD_TABLE_NAME} (name) VALUES (?)'
            params = (name,)

            if insert_dt and update_dt:
                query = f'INSERT INTO {THREAD_TABLE_NAME} (name, insert_dt, update_dt) VALUES (?, ?, ?)'
                params = (name, insert_dt, update_dt)
            elif insert_dt:
                query = f'INSERT INTO {THREAD_TABLE_NAME} (name, insert_dt) VALUES (?, ?)'
                params = (name, insert_dt)

            # Insert a row into the table
            self.__c.execute(query, params)
            new_id = self.__c.lastrowid
            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updateThread(self, id, name):
        try:
            self.__c.execute(f'UPDATE {THREAD_TABLE_NAME} SET name=(?) WHERE id={id}', (name,))
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def deleteThread(self, id=None):
        try:
            query = f'DELETE FROM {THREAD_TABLE_NAME}'
            if id:
                query += f' WHERE id = {id}'
            self.__c.execute(query)
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createMessageTrigger(self, insert_trigger=True, update_trigger=True, delete_trigger=True):
        """
        Create message trigger
        """
        if insert_trigger:
            # Create insert trigger
            self.__c.execute(f'''
                CREATE TRIGGER {THREAD_MESSAGE_INSERTED_TR_NAME}
                AFTER INSERT ON {MESSAGE_TABLE_NAME}
                BEGIN
                  UPDATE {THREAD_TABLE_NAME} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.thread_id;
                END
            ''')

        if update_trigger:
            # Create update trigger
            self.__c.execute(f'''
                CREATE TRIGGER {THREAD_MESSAGE_UPDATED_TR_NAME}
                AFTER UPDATE ON {MESSAGE_TABLE_NAME}
                BEGIN
                  UPDATE {THREAD_TABLE_NAME} SET update_dt = CURRENT_TIMESTAMP WHERE id = NEW.thread_id;
                END
            ''')

        if delete_trigger:
            # Create delete trigger
            self.__c.execute(f'''
                CREATE TRIGGER {THREAD_MESSAGE_DELETED_TR_NAME}
                AFTER DELETE ON {MESSAGE_TABLE_NAME}
                BEGIN
                  UPDATE {THREAD_TABLE_NAME} SET update_dt = CURRENT_TIMESTAMP WHERE id = OLD.thread_id;
                END
            ''')

        # Commit the transaction
        self.__conn.commit()

    def __createMessage(self):
        """
        Create message table
        """
        try:
            # Check if the table exists
            self.__c.execute(
                f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{MESSAGE_TABLE_NAME}'")
            if self.__c.fetchone()[0] == 1:
                # TODO WILL_BE_REMOVED_AFTER v1.6.0
                # If there is no is_g4f column, add it
                self.__c.execute(f"PRAGMA table_info({MESSAGE_TABLE_NAME})")
                columns = self.__c.fetchall()
                if 'is_g4f' not in [col[1] for col in columns]:
                    # Add is_g4f, provider to the table
                    self.__c.execute(f'ALTER TABLE {MESSAGE_TABLE_NAME} ADD COLUMN is_g4f INT DEFAULT 0')

                # If there is no provider column, add it
                if 'provider' not in [col[1] for col in columns]:
                    self.__c.execute(f"ALTER TABLE {MESSAGE_TABLE_NAME} ADD COLUMN provider VARCHAR(255) DEFAULT ''")
            else:
                # Create message table and triggers
                self.__c.execute(f'''CREATE TABLE {MESSAGE_TABLE_NAME}
                             (id INTEGER PRIMARY KEY,
                              thread_id INTEGER,
                              role VARCHAR(255),
                              content TEXT,
                              finish_reason VARCHAR(255),
                              model VARCHAR(255),
                              prompt_tokens INTEGER,
                              completion_tokens INTEGER,
                              total_tokens INTEGER,
                              favorite INTEGER DEFAULT 0,
                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                              favorite_set_date DATETIME,
                              is_json_response_available INT DEFAULT 0,
                              is_g4f INT DEFAULT 0,
                              provider VARCHAR(255),
                              FOREIGN KEY (thread_id) REFERENCES {THREAD_TABLE_NAME}(id)
                              ON DELETE CASCADE)''')

                self.__createMessageTrigger()
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def selectCertainThreadMessagesRaw(self, thread_id, content_to_select=None):
        """
        This is for selecting all messages in a thread with a specific thread_id.
        The format of the result is a list of sqlite Rows.
        """
        # Begin the query with the thread_id filter
        query = f'SELECT * FROM {MESSAGE_TABLE_NAME} WHERE thread_id = ?'
        params = [thread_id]  # Start the parameter list with the thread_id

        # If content_to_select is provided, append to the query
        if content_to_select:
            query += ' AND LOWER(content) LIKE LOWER(?)'  # Modify for case-insensitive
            params.append(f'%{content_to_select}%')  # Use parameterized placeholder

        # Execute the query with parameters
        self.__c.execute(query, params)

        # Fetch all results and return
        return self.__c.fetchall()

    def selectCertainThreadMessages(self, thread_id, content_to_select=None) -> List[ChatMessageContainer]:
        """
        This is for selecting all messages in a thread with a specific thread_id.
        The format of the result is a list of ChatMessageContainer.
        """
        result = [ChatMessageContainer(**elem) for elem in self.selectCertainThreadMessagesRaw(thread_id, content_to_select=content_to_select)]
        return result

    def selectAllContentOfThread(self, content_to_select=None):
        """
        This is for selecting all messages in all threads which include the content_to_select.
        """
        arr = []
        for _id in [conv[0] for conv in self.selectAllThread()]:
            result = self.selectCertainThreadMessages(_id, content_to_select)
            if result:
                arr.append((_id, result))
        return arr

    def insertMessage(self, arg: ChatMessageContainer, deactivate_trigger=False):
        try:
            if deactivate_trigger:
             # Remove the trigger
                self.__c.execute(f'DROP TRIGGER {THREAD_MESSAGE_INSERTED_TR_NAME}')
            excludes = ['id', 'update_dt', 'insert_dt']
            insert_query = arg.create_insert_query(table_name=MESSAGE_TABLE_NAME, excludes=excludes)
            self.__c.execute(insert_query, arg.get_values_for_insert(excludes=excludes))
            new_id = self.__c.lastrowid
            if deactivate_trigger:
                # Create the trigger
                self.__createMessageTrigger(insert_trigger=True, update_trigger=False, delete_trigger=False)

            # Commit the transaction
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def updateMessage(self, id, favorite):
        """
        Update message favorite
        """
        try:
            current_date = datetime.now().strftime(DEFAULT_DATETIME_FORMAT)
            self.__c.execute(f'''
                            UPDATE {MESSAGE_TABLE_NAME} 
                            SET favorite = ?,
                                favorite_set_date = CASE 
                                                      WHEN ? = 1 THEN ? 
                                                      ELSE NULL 
                                                    END 
                            WHERE id = ?
                        ''', (favorite, favorite, current_date, id))
            self.__conn.commit()
            return current_date
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def __createImage(self):
        try:
            # Check if the table exists
            self.__c.execute(f"SELECT count(*) FROM sqlite_master WHERE type='table' AND name='{IMAGE_TABLE_NAME}'")
            if self.__c.fetchone()[0] == 1:
                pass
            else:
                self.__c.execute(f'''CREATE TABLE {IMAGE_TABLE_NAME}
                             (id INTEGER PRIMARY KEY,
                              model VARCHAR(255),
                              prompt TEXT,
                              n INT,
                              quality VARCHAR(255),
                              data BLOB,
                              style VARCHAR(255),
                              revised_prompt TEXT,
                              width INT,
                              height INT,
                              negative_prompt TEXT,
                              update_dt DATETIME DEFAULT CURRENT_TIMESTAMP,
                              insert_dt DATETIME DEFAULT CURRENT_TIMESTAMP)''')
                # Commit the transaction
                self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred while creating the table: {e}")
            raise

    def insertImage(self, arg: ImagePromptContainer):
        try:
            excludes = ['id', 'insert_dt', 'update_dt']
            query = arg.create_insert_query(IMAGE_TABLE_NAME, excludes)
            values = arg.get_values_for_insert(excludes)
            self.__c.execute(query, values)
            new_id = self.__c.lastrowid
            self.__conn.commit()
            return new_id
        except sqlite3.Error as e:
            print(f"An error occurred..")
            raise

    def selectImage(self):
        try:
            self.__c.execute(f'SELECT * FROM {IMAGE_TABLE_NAME}')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectCertainImage(self, id):
        try:
            self.__c.execute(f'SELECT * FROM {IMAGE_TABLE_NAME} WHERE id={id}')
            return self.__c.fetchone()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def removeImage(self, id=None):
        try:
            query = f'DELETE FROM {IMAGE_TABLE_NAME}'
            if id:
                query += f' WHERE id = {id}'
            self.__c.execute(query)
            self.__conn.commit()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def selectFavorite(self):
        try:
            self.__c.execute(f'SELECT * FROM {MESSAGE_TABLE_NAME} WHERE favorite=1 order by favorite_set_date')
            return self.__c.fetchall()
        except sqlite3.Error as e:
            print(f"An error occurred: {e}")
            raise

    def export(self, ids, filename):
        # Get the records of the threads of the given ids
        thread_records = self.selectAllThread(ids)
        data = [dict(record) for record in thread_records]
        # Convert it into dictionary
        for d in data:
            d['messages'] = list(map(lambda x: x.__dict__, self.selectCertainThreadMessages(d['id'])))

        # Save the JSON
        with open(filename, 'w') as f:
            json.dump(data, f)

    def getCursor(self):
        return self.__c

    def close(self):
        self.__conn.close()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Close the connection
        self.__conn.close()



import json

from PySide6.QtCore import QLocale

from pyqt_openai import DEFAULT_LANGUAGE, LANGUAGE_FILE, LANGUAGE_DICT


class WordsDict(dict):
    """
    Only used for release version
    to prevent KeyError
    """
    def __missing__(self, key):
        return key


class LangClass:
    """
    LangClass is the class that manages the language of the application.
    It reads the language file and sets the language.
    """
    TRANSLATIONS = WordsDict()

    @classmethod
    def lang_changed(cls, lang=None):
        with open(LANGUAGE_FILE, 'r', encoding='utf-8') as file:
            translations_data = json.load(file)

        if not lang:
            language = QLocale.system().name()
            if language not in translations_data:
                language = DEFAULT_LANGUAGE  # Default language
        else:
            language = LANGUAGE_DICT[lang]

        cls.TRANSLATIONS = WordsDict(translations_data[language])

        for k, v in LANGUAGE_DICT.items():
            if v == language:
                return k


import configparser
import os
import shutil

import yaml

from pyqt_openai import CONFIG_DATA, INI_FILE_NAME, get_config_directory, ROOT_DIR, SRC_DIR

_config_cache = None


def parse_value(value):
    # Boolean conversion
    if value.lower() in ('true', 'false'):
        return value.lower() == 'true'
    # Numeric conversion
    try:
        return int(value)
    except ValueError:
        try:
            return float(value)
        except ValueError:
            pass
    # Default: return the value as is (string)
    return value

def convert_list(value):
    # Convert comma-separated string to list
    return [item.strip() for item in value.split(',')]

def init_yaml():
    yaml_data = CONFIG_DATA

    config_dir = get_config_directory()
    config_path = os.path.join(config_dir, INI_FILE_NAME)

    if not os.path.exists(config_path):
        # Save as YAML file
        with open(config_path, 'w') as yaml_file:
            yaml.dump(yaml_data, yaml_file, default_flow_style=False)
    else:
        with open(config_path, 'r') as yaml_file:
            prev_yaml_data = yaml.safe_load(yaml_file)
        # Add new keys
        for section, values in yaml_data.items():
            if section not in prev_yaml_data:
                prev_yaml_data[section] = values
            else:
                for key, value in values.items():
                    if key not in prev_yaml_data[section]:
                        prev_yaml_data[section][key] = value
        # Save as YAML file
        with open(config_path, 'w') as yaml_file:
            yaml.dump(prev_yaml_data, yaml_file, default_flow_style=False)


class ConfigManager:
    def __init__(self, yaml_file=INI_FILE_NAME):
        self.yaml_file = yaml_file
        self.config = self._load_yaml()

    def _load_yaml(self):
        with open(self.yaml_file, 'r') as file:
            return yaml.safe_load(file)

    def _save_yaml(self):
        with open(self.yaml_file, 'w') as file:
            yaml.safe_dump(self.config, file)

    # Getter methods
    def get_dalle(self):
        return self.config.get('DALLE', {})

    def get_general(self):
        return self.config.get('General', {})

    def get_replicate(self):
        return self.config.get('REPLICATE', {})

    def get_dalle_property(self, key):
        return self.config.get('DALLE', {}).get(key)

    def get_general_property(self, key):
        return self.config.get('General', {}).get(key)

    def get_replicate_property(self, key):
        return self.config.get('REPLICATE', {}).get(key)

    # Setter methods
    def set_dalle_property(self, key, value):
        if 'DALLE' not in self.config:
            self.config['DALLE'] = {}
        self.config['DALLE'][key] = value
        self._save_yaml()

    def set_general_property(self, key, value):
        if 'General' not in self.config:
            self.config['General'] = {}
        self.config['General'][key] = value
        self._save_yaml()

    def set_replicate_property(self, key, value):
        if 'REPLICATE' not in self.config:
            self.config['REPLICATE'] = {}
        self.config['REPLICATE'][key] = value
        self._save_yaml()

def update_api_key(yaml_file_path):
    with open(yaml_file_path, 'r') as file:
        data = yaml.safe_load(file)

    if 'General' in data and 'API_KEY' in data['General']:
        data['General']['OPENAI_API_KEY'] = data['General'].pop('API_KEY')

    with open(yaml_file_path, 'w') as file:
        yaml.safe_dump(data, file)

init_yaml()

# TODO WILL_BE_REMOVED_AFTER v1.5.0
update_api_key(INI_FILE_NAME)

CONFIG_MANAGER = ConfigManager()


from PySide6.QtCore import Qt, Signal
from PySide6.QtWidgets import QDialog, QLabel, QPushButton, QVBoxLayout, QHBoxLayout, QWidget, \
    QCheckBox, QSpacerItem, QSizePolicy

from pyqt_openai.lang.translations import LangClass
from pyqt_openai.util.script import getSeparator


class DoNotAskAgainDialog(QDialog):
    doNotAskAgainChanged = Signal(bool)

    def __init__(self, do_not_ask_again: bool = False, parent=None):
        super().__init__(parent)
        do_not_ask_again_message = LangClass.TRANSLATIONS["Would you like to exit the application? If you won't, it will be running in the background."]
        do_not_ask_again_checkbox_message = LangClass.TRANSLATIONS['Do not ask again']
        self.__initVal(do_not_ask_again, do_not_ask_again_message, do_not_ask_again_checkbox_message)
        self.__initUi()

    def __initVal(self, do_not_ask_again, do_not_ask_again_message, do_not_ask_again_checkbox_message):
        self.__is_cancel = False
        self.__do_not_ask_again = do_not_ask_again
        self.__do_not_ask_again_message = do_not_ask_again_message
        self.__do_not_ask_again_checkbox_message = do_not_ask_again_checkbox_message

    def __initUi(self):
        self.setWindowTitle(LangClass.TRANSLATIONS["Exit"])
        self.setModal(True)
        self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowCloseButtonHint)

        self.label = QLabel(self.__do_not_ask_again_message)
        self.label.setAlignment(Qt.AlignmentFlag.AlignCenter)

        self.yesButton = QPushButton(LangClass.TRANSLATIONS["Yes"])
        self.yesButton.clicked.connect(self.accept)

        self.noButton = QPushButton(LangClass.TRANSLATIONS["No"])
        self.noButton.clicked.connect(self.reject)

        self.cancelButton = QPushButton(LangClass.TRANSLATIONS["Cancel"])
        self.cancelButton.clicked.connect(self.__cancel)

        self.doNotAskAgainCheckBox = QCheckBox(self.__do_not_ask_again_checkbox_message)
        self.doNotAskAgainCheckBox.setChecked(self.__do_not_ask_again)
        self.doNotAskAgainCheckBox.stateChanged.connect(self.__onCheckBoxStateChanged)

        sep = getSeparator('horizontal')

        lay = QHBoxLayout()
        lay.addWidget(self.doNotAskAgainCheckBox)
        lay.addSpacerItem(QSpacerItem(10, 10, QSizePolicy.Policy.MinimumExpanding))
        lay.addWidget(self.yesButton)
        lay.addWidget(self.noButton)
        lay.addWidget(self.cancelButton)
        lay.setAlignment(Qt.AlignmentFlag.AlignRight)
        lay.setContentsMargins(0,0,0,0)
        btnWidget = QWidget()
        btnWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(self.label)
        lay.addWidget(sep)
        lay.addWidget(btnWidget)

        self.setLayout(lay)

    def __cancel(self):
        self.__is_cancel = True
        self.reject()

    def __onCheckBoxStateChanged(self, state):
        self.__do_not_ask_again = state == 2
        self.doNotAskAgainChanged.emit(self.__do_not_ask_again)

    def isCancel(self):
        return self.__is_cancel


"""
This file is used to store the data classes that are used throughout the application.
"""

from dataclasses import dataclass, fields, field
from typing import List

from pyqt_openai import DB_FILE_NAME, DEFAULT_FONT_SIZE, DEFAULT_FONT_FAMILY, DEFAULT_USER_IMAGE_PATH, \
    DEFAULT_AI_IMAGE_PATH, MAXIMUM_MESSAGES_IN_PARAMETER, \
    WHISPER_TTS_DEFAULT_VOICE, WHISPER_TTS_DEFAULT_SPEED
from pyqt_openai.lang.translations import LangClass


@dataclass
class Container:
    def __init__(self, **kwargs):
        """
        You don't have to call this if you want to use default class variables
        """
        for k in self.__annotations__:
            setattr(self, k, kwargs.get(k, getattr(self, k, "")))
        for key, value in kwargs.items():
            if key in self.__annotations__:
                setattr(self, key, value)

    @classmethod
    def get_keys(cls, excludes: list = None):
        """
        Function that returns the keys of the target data type as a list.
        Exclude the keys in the "excludes" list.
        """
        if excludes is None:
            excludes = []
        arr = [field.name for field in fields(cls)]
        for exclude in excludes:
            if exclude in arr:
                arr.remove(exclude)
        return arr

    def get_values_for_insert(self, excludes: list = None):
        """
        Function that returns the values of the target data type as a list.
        """
        if excludes is None:
            excludes = []
        arr = [getattr(self, key) for key in self.get_keys(excludes)]
        return arr

    def get_items(self, excludes: list = None):
        """
        Function that returns the items of the target data type as a list.
        """
        if excludes is None:
            excludes = []
        return {key: getattr(self, key) for key in self.get_keys(excludes)}.items()

    def create_insert_query(self, table_name: str, excludes: list = None):
        if excludes is None:
            excludes = []
        """
        Function to dynamically generate an SQLite insert statement.
        Takes the table name as a parameter.
        """
        field_names = self.get_keys(excludes)
        columns = ', '.join(field_names)
        placeholders = ', '.join(['?' for _ in field_names])
        query = f'INSERT INTO {table_name} ({columns}) VALUES ({placeholders})'
        return query

@dataclass
class ChatThreadContainer(Container):
    id: str = ""
    name: str = ""
    insert_dt: str = ""
    update_dt: str = ""

@dataclass
class ChatMessageContainer(Container):
    id: str = ""
    thread_id: str = ""
    role: str = ""
    content: str = ""
    insert_dt: str = ""
    update_dt: str = ""
    finish_reason: str = ""
    model: str = ""
    prompt_tokens: str = ""
    completion_tokens: str = ""
    total_tokens: str = ""
    favorite: int = 0
    favorite_set_date: str = ""
    is_json_response_available: str = 0
    is_g4f: int = 0
    provider: str = ""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class ImagePromptContainer(Container):
    id: str = ""
    model: str = ""
    width: str = ""
    height: str = ""
    prompt: str = ""
    negative_prompt: str = ""
    n: str = ""
    quality: str = ""
    data: str = ""
    style: str = ""
    revised_prompt: str = ""
    update_dt: str = ""
    insert_dt: str = ""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

@dataclass
class SettingsParamsContainer(Container):
    lang: str = LangClass.lang_changed()
    db: str = DB_FILE_NAME
    do_not_ask_again: bool = False
    notify_finish: bool = True
    show_toolbar: bool = True
    show_secondary_toolbar: bool = True
    chat_column_to_show: List[str] = field(default_factory=ChatThreadContainer.get_keys)
    image_column_to_show: List[str] = field(default_factory=ImagePromptContainer.get_keys)
    maximum_messages_in_parameter: int = MAXIMUM_MESSAGES_IN_PARAMETER
    show_as_markdown: bool = True
    apply_user_defined_styles: bool = False
    run_at_startup: bool = True
    voice: str = WHISPER_TTS_DEFAULT_VOICE
    voice_speed: int = WHISPER_TTS_DEFAULT_SPEED

@dataclass
class CustomizeParamsContainer(Container):
    background_image: str = ''
    user_image: str = DEFAULT_USER_IMAGE_PATH
    ai_image: str = DEFAULT_AI_IMAGE_PATH
    font_size: int = DEFAULT_FONT_SIZE
    font_family: str = DEFAULT_FONT_FAMILY

@dataclass
class PromptGroupContainer(Container):
    id: str = ""
    name: str = ""
    insert_dt: str = ""
    update_dt: str = ""
    prompt_type: str = ""

@dataclass
class PromptEntryContainer(Container):
    id: str = ""
    group_id: str = ""
    name: str = ""
    content: str = ""
    insert_dt: str = ""
    update_dt: str = ""


from PySide6.QtCore import Signal, Qt, QThread
from PySide6.QtGui import QFontDatabase, QFont
from PySide6.QtWidgets import QListWidget, QWidget, QVBoxLayout, QLabel, QLineEdit, QListWidgetItem
from PySide6.QtWidgets import QSizePolicy, \
    QTextEdit, QHBoxLayout

from pyqt_openai import DEFAULT_FONT_FAMILY
from pyqt_openai.lang.translations import LangClass


class FontLoaderThread(QThread):
    fonts_loaded = Signal(list)
    afterFinished = Signal(QFont)

    def __init__(self, font: QFont):
        super().__init__()
        self.font = font

    def run(self):
        fm = QFontDatabase.families(QFontDatabase.Any)
        self.fonts_loaded.emit(fm)
        self.afterFinished.emit(self.font)


class SizeWidget(QWidget):
    sizeItemChanged = Signal(int)

    def __init__(self, font: QFont, parent=None):
        super().__init__(parent)
        self.__initUi(font=font)

    def __initUi(self, font: QFont):
        self.__sizeLineEdit = QLineEdit()
        self.__sizeLineEdit.textEdited.connect(self.__textEdited)

        self.__sizeListWidget = QListWidget()
        self.__initSizes(font=font)
        self.__sizeListWidget.itemSelectionChanged.connect(self.__sizeItemChanged)

        lay = QVBoxLayout()
        lay.addWidget(self.__sizeLineEdit)
        lay.addWidget(self.__sizeListWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)

        sizeBottomWidget = QWidget()
        sizeBottomWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(QLabel(LangClass.TRANSLATIONS['Size']))
        lay.addWidget(sizeBottomWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(5)

        self.setLayout(lay)

    def __initSizes(self, font: QFont):
        self.__initSizesList(font=font)
        self.setCurrentSize(font=font)

    def __initSizesList(self, font: QFont):
        font_name = font.family()
        style_name = QFontDatabase.styles(font_name)
        # In case of font is not in the font list
        if style_name:
            pass
        else:
            font_name = 'Arial'
        sizes = QFontDatabase.pointSizes(font_name)
        sizes = list(map(str, sizes))
        self.__sizeListWidget.addItems(sizes)

    def setCurrentSize(self, font: QFont):
        items = self.__sizeListWidget.findItems(str(font.pointSize()), Qt.MatchFlag.MatchFixedString)
        item = QListWidgetItem()
        if items:
            item = items[0]
        else:
            item = self.__sizeListWidget.item(0)
        if item:
            self.__sizeListWidget.setCurrentItem(item)
            size_text = item.text()
            self.__sizeLineEdit.setText(size_text)
        else:
            item = QListWidgetItem('10')
            self.__sizeListWidget.setCurrentItem(item)
            size_text = item.text()
            self.__sizeLineEdit.setText(size_text)

    def __textEdited(self):
        size_text = self.__sizeLineEdit.text()
        items = self.__sizeListWidget.findItems(size_text, Qt.MatchFlag.MatchFixedString)
        if items:
            self.__sizeListWidget.setCurrentItem(items[0])
        self.sizeItemChanged.emit(int(size_text))

    def __sizeItemChanged(self):
        size_text = self.__sizeListWidget.currentItem().text()
        self.sizeItemChanged.emit(int(size_text))
        self.__sizeLineEdit.setText(size_text)

    def setSizes(self, sizes, prev_size=10):
        sizes = list(map(str, sizes))
        self.__sizeListWidget.clear()
        self.__sizeListWidget.addItems(sizes)
        items = self.__sizeListWidget.findItems(str(prev_size), Qt.MatchFlag.MatchFixedString)
        if len(items) > 0:
            item = items[0]
            self.__sizeListWidget.setCurrentItem(item)
            self.__sizeLineEdit.setText(item.text())
        else:
            self.__sizeLineEdit.setText(str(prev_size))

    def getSize(self):
        return self.__sizeListWidget.currentItem().text() if self.__sizeListWidget.currentItem() else 10

class FontItemWidget(QWidget):
    fontItemChanged = Signal(str, list, list)

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.__font_families = []
        self.__initUi(font=font)

    def __initUi(self, font: QFont):
        self.__fontLineEdit = QLineEdit()
        self.__fontLineEdit.textEdited.connect(self.__textEdited)

        self.__fontListWidget = QListWidget()
        self.__initFonts(font)
        self.__fontListWidget.itemSelectionChanged.connect(self.__fontItemChanged)

        lay = QVBoxLayout()
        lay.addWidget(self.__fontLineEdit)
        lay.addWidget(self.__fontListWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)

        fontBottomWidget = QWidget()
        fontBottomWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(QLabel(LangClass.TRANSLATIONS['Font']))
        lay.addWidget(fontBottomWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(5)

        self.setLayout(lay)

    def __initFonts(self, font: QFont):
        self.loader_thread = FontLoaderThread(font)
        self.loader_thread.fonts_loaded.connect(self.__onFontsLoaded)
        self.loader_thread.start()
        self.loader_thread.afterFinished.connect(self.setCurrentFont)

    def __onFontsLoaded(self, fm):
        self.__font_families.extend(fm)
        # Set each item to each font family
        for f in fm:
            item = QListWidgetItem(f)
            # FIXME This makes the font list widget too slow
            # item.setFont(QFont(f))
            self.__fontListWidget.addItem(item)

    def setCurrentFont(self, font: QFont):
        items = self.__fontListWidget.findItems(font.family(), Qt.MatchFlag.MatchFixedString)
        item = QListWidgetItem()
        if items:
            item = items[0]
        else:
            item = self.__fontListWidget.item(0)
        self.__fontListWidget.setCurrentItem(item)
        font_name = item.text()
        self.__fontLineEdit.setText(font_name)

    def __fontItemChanged(self):
        font_name = self.__fontListWidget.currentItem().text()
        self.__fontLineEdit.setText(font_name)
        styles = QFontDatabase.styles(font_name)
        pointSizes = QFontDatabase.pointSizes(font_name, QFontDatabase.styles(font_name)[0])
        self.fontItemChanged.emit(font_name, styles, pointSizes)

    def __textEdited(self):
        self.__fontListWidget.clear()
        text = self.__fontLineEdit.text()
        if text.strip() != '':
            match_families = []
            for family in self.__font_families:
                if family.startswith(text):
                    match_families.append(family)
            if match_families:
                self.__fontListWidget.addItems(match_families)
            else:
                pass
        else:
            self.__fontListWidget.addItems(self.__font_families)

    def getFontFamily(self):
        item = self.__fontListWidget.currentItem()
        if item:
            return item.text()
        else:
            return DEFAULT_FONT_FAMILY


class FontWidget(QWidget):
    fontChanged = Signal(QFont)

    def __init__(self, font, parent=None):
        super().__init__(parent)
        self.__current_font = font
        self.__initUi(font=font)

    def __initUi(self, font: QFont):
        self.__previewTextEdit = QTextEdit(self)
        self.__previewTextEdit.textChanged.connect(self.__textChanged)

        self.__fontItemWidget = FontItemWidget(font)
        self.__fontItemWidget.fontItemChanged.connect(self.__fontItemChangedExec)

        self.__sizeWidget = SizeWidget(font)
        self.__sizeWidget.sizeItemChanged.connect(self.__sizeItemChangedExec)

        self.__initPreviewTextEdit()

        lay = QHBoxLayout()
        lay.addWidget(self.__fontItemWidget)
        lay.addWidget(self.__sizeWidget)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(0)

        topWidget = QWidget()
        topWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(QLabel(LangClass.TRANSLATIONS['Preview']))
        lay.addWidget(self.__previewTextEdit)
        lay.setContentsMargins(0, 0, 0, 0)
        lay.setSpacing(5)

        bottomWidget = QWidget()
        bottomWidget.setLayout(lay)
        bottomWidget.setSizePolicy(QSizePolicy.MinimumExpanding, QSizePolicy.Preferred)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(bottomWidget)
        self.setLayout(lay)

    def __initPreviewTextEdit(self):
        font_family = self.__fontItemWidget.getFontFamily()
        font_size = self.__sizeWidget.getSize()
        font = self.__previewTextEdit.currentFont()
        font.setFamily(font_family)
        font.setPointSize(int(font_size))
        self.__previewTextEdit.setCurrentFont(font)
        self.__previewTextEdit.setText(LangClass.TRANSLATIONS['Sample'])

    def __sizeItemChangedExec(self, size):
        self.__previewTextEdit.selectAll()
        font = self.__previewTextEdit.currentFont()
        font.setPointSize(size)
        self.__previewTextEdit.setCurrentFont(font)

        self.__current_font = font
        self.fontChanged.emit(self.__current_font)

    def __fontItemChangedExec(self, font_text, styles, sizes):
        self.__previewTextEdit.selectAll()
        font = self.__previewTextEdit.currentFont()
        prev_size = font.pointSize()

        font.setFamily(font_text)

        sizes = list(filter(lambda x: x <= 20 and x >= 8, sizes))

        if prev_size in sizes:
            self.__sizeWidget.setSizes(sizes, prev_size)
            font.setPointSize(prev_size)
        else:
            self.__sizeWidget.setSizes(sizes, prev_size)

        self.__previewTextEdit.setCurrentFont(font)
        self.__current_font = font
        self.fontChanged.emit(self.__current_font)

    def getFont(self):
        return self.__previewTextEdit.currentFont()

    def setCurrentFont(self, font):
        self.__fontItemWidget.setCurrentFont(font=font)
        self.__sizeWidget.setCurrentSize(font=font)

        self.__previewTextEdit.setCurrentFont(font)
        self.__current_font = font
        self.fontChanged.emit(self.__current_font)

    def __textChanged(self):
        text = self.__previewTextEdit.toPlainText()
        if text.strip() != '':
            pass
        else:
            self.__previewTextEdit.setCurrentFont(self.__current_font)


import datetime

from PySide6.QtCore import Qt
from PySide6.QtGui import QPixmap
from PySide6.QtWidgets import QDialog, QPushButton, QHBoxLayout, QWidget, QVBoxLayout, QLabel

import pyqt_openai
from pyqt_openai import DEFAULT_APP_ICON, LICENSE_URL, DEFAULT_APP_NAME, CONTACT
from pyqt_openai.lang.translations import LangClass
from pyqt_openai.widgets.linkLabel import LinkLabel


class AboutDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.__initUi()

    def __initUi(self):
        self.setWindowTitle(LangClass.TRANSLATIONS['About'])
        self.setWindowFlags(Qt.WindowType.Window | Qt.WindowType.WindowCloseButtonHint)

        self.__okBtn = QPushButton(LangClass.TRANSLATIONS['OK'])
        self.__okBtn.clicked.connect(self.accept)

        p = QPixmap(DEFAULT_APP_ICON)
        logoLbl = QLabel()
        logoLbl.setPixmap(p)

        descWidget1 = QLabel()
        descWidget1.setText(f'''
        <h1>{DEFAULT_APP_NAME}</h1>
        Software Version {pyqt_openai.__version__}<br/><br/>
        © 2023 {datetime.datetime.now().year}. Used under the {pyqt_openai.LICENSE} License.<br/>
        Copyright (c) {datetime.datetime.now().year} {pyqt_openai.__author__}<br/>
        ''')

        descWidget2 = LinkLabel()
        descWidget2.setText(LangClass.TRANSLATIONS['Read MIT License Full Text'])
        descWidget2.setUrl(LICENSE_URL)

        descWidget3 = QLabel()
        descWidget3.setText(f'''
        <br/><br/>Contact: {CONTACT}<br/>
        <p>Powered by PySide6</p>
        <p>Powered by GPT4Free</p>
        ''')

        descWidget1.setAlignment(Qt.AlignmentFlag.AlignTop)
        descWidget2.setAlignment(Qt.AlignmentFlag.AlignTop)
        descWidget3.setAlignment(Qt.AlignmentFlag.AlignTop)

        lay = QVBoxLayout()
        lay.addWidget(descWidget1)
        lay.addWidget(descWidget2)
        lay.addWidget(descWidget3)
        lay.setAlignment(Qt.AlignmentFlag.AlignTop)
        lay.setContentsMargins(0, 0, 0, 0)

        rightWidget = QWidget()
        rightWidget.setLayout(lay)

        lay = QHBoxLayout()
        lay.addWidget(logoLbl)
        lay.addWidget(rightWidget)

        topWidget = QWidget()
        topWidget.setLayout(lay)

        cancelBtn = QPushButton(LangClass.TRANSLATIONS['Cancel'])
        cancelBtn.clicked.connect(self.close)

        lay = QHBoxLayout()
        lay.addWidget(self.__okBtn)
        lay.addWidget(cancelBtn)
        lay.setAlignment(Qt.AlignmentFlag.AlignRight)
        lay.setContentsMargins(0, 0, 0, 0)

        okCancelWidget = QWidget()
        okCancelWidget.setLayout(lay)

        lay = QVBoxLayout()
        lay.addWidget(topWidget)
        lay.addWidget(okCancelWidget)

        self.setLayout(lay)

